import errno
import webbrowser
from http.server import HTTPServer, SimpleHTTPRequestHandler
from unittest import mock

import pytest

from briefcase.console import Console, Log
from briefcase.exceptions import BriefcaseCommandError
from briefcase.platforms.web.static import (
    HTTPHandler,
    LocalHTTPServer,
    StaticWebRunCommand,
)


# OSError doesn't expose errno in the constructor; create some
# custom exceptions that mock common connection errors.
class ErrnoError(OSError):
    def __init__(self, errno):
        super().__init__()
        self.errno = errno


@pytest.fixture
def run_command(tmp_path):
    command = StaticWebRunCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.data_path = tmp_path / "briefcase"
    return command


def test_default_options(run_command):
    """The default options are as expected."""
    options, overrides = run_command.parse_options([])

    assert options == {
        "appname": None,
        "update": False,
        "update_requirements": False,
        "update_resources": False,
        "update_support": False,
        "update_stub": False,
        "no_update": False,
        "test_mode": False,
        "passthrough": [],
        "host": "localhost",
        "port": 8080,
        "open_browser": True,
    }
    assert overrides == {}


def test_options(run_command):
    """The extra options can be parsed."""
    options, overrides = run_command.parse_options(
        ["--host", "myhost", "--port", "1234", "--no-browser"]
    )

    assert options == {
        "appname": None,
        "update": False,
        "update_requirements": False,
        "update_resources": False,
        "update_support": False,
        "update_stub": False,
        "no_update": False,
        "test_mode": False,
        "passthrough": [],
        "host": "myhost",
        "port": 1234,
        "open_browser": False,
    }
    assert overrides == {}


def test_run(monkeypatch, run_command, first_app_built):
    """A static web app can be launched as a server."""
    # Mock server creation
    mock_server_init = mock.MagicMock(spec_set=HTTPServer)
    monkeypatch.setattr(HTTPServer, "__init__", mock_server_init)

    # Mock the socket name returned by the server.
    socket = mock.MagicMock()
    socket.getsockname.return_value = ("127.0.0.1", "8080")
    LocalHTTPServer.socket = socket

    # Mock server execution, raising a user exit.
    mock_serve_forever = mock.MagicMock(side_effect=KeyboardInterrupt())
    monkeypatch.setattr(HTTPServer, "serve_forever", mock_serve_forever)

    # Mock shutdown
    mock_shutdown = mock.MagicMock()
    monkeypatch.setattr(HTTPServer, "shutdown", mock_shutdown)

    # Mock server close
    mock_server_close = mock.MagicMock()
    monkeypatch.setattr(HTTPServer, "server_close", mock_server_close)

    # Mock the webbrowser
    mock_open_new_tab = mock.MagicMock()
    monkeypatch.setattr(webbrowser, "open_new_tab", mock_open_new_tab)

    # Run the app
    run_command.run_app(
        first_app_built,
        test_mode=False,
        passthrough=[],
        host="localhost",
        port=8080,
        open_browser=True,
    )

    # The browser was opened
    mock_open_new_tab.assert_called_once_with("http://127.0.0.1:8080")

    # The server was started
    mock_serve_forever.assert_called_once_with()

    # The webserver was shutdown.
    mock_shutdown.assert_called_once_with()

    # The webserver was closed.
    mock_server_close.assert_called_once_with()


@pytest.mark.parametrize(
    "exception",
    [
        ErrnoError(errno.EADDRINUSE),
        ErrnoError(errno.ENOSR),
    ],
)
def test_run_with_fallback_port(
    monkeypatch,
    run_command,
    first_app_built,
    exception,
    capsys,
):
    """A static web app can be launched as a server even when the requested port is
    already in use."""
    # Mock server creation that first errors on port, then connects with port 0
    mock_server_init = mock.MagicMock(side_effect=[exception, HTTPServer])
    monkeypatch.setattr(HTTPServer, "__init__", mock_server_init)

    # Mock the socket name returned by the server.
    # This value has been auto-selected by the server.
    socket = mock.MagicMock()
    socket.getsockname.return_value = ("127.0.0.1", "12345")
    LocalHTTPServer.socket = socket

    # Mock server execution, raising a user exit.
    mock_serve_forever = mock.MagicMock(side_effect=KeyboardInterrupt())
    monkeypatch.setattr(HTTPServer, "serve_forever", mock_serve_forever)

    # Mock shutdown
    mock_shutdown = mock.MagicMock()
    monkeypatch.setattr(HTTPServer, "shutdown", mock_shutdown)

    # Mock server close
    mock_server_close = mock.MagicMock()
    monkeypatch.setattr(HTTPServer, "server_close", mock_server_close)

    # Mock the webbrowser
    mock_open_new_tab = mock.MagicMock()
    monkeypatch.setattr(webbrowser, "open_new_tab", mock_open_new_tab)

    # Run the app
    run_command.run_app(
        first_app_built,
        test_mode=False,
        passthrough=[],
        host="localhost",
        port=8080,
        open_browser=True,
    )

    # User is warned a system-allocated port is being used
    assert "Using a system-allocated port since port 8080" in capsys.readouterr().out

    # The browser was opened
    mock_open_new_tab.assert_called_once_with("http://127.0.0.1:12345")

    # The server was started
    mock_serve_forever.assert_called_once_with()

    # The webserver was shutdown.
    mock_shutdown.assert_called_once_with()

    # The webserver was closed.
    mock_server_close.assert_called_once_with()


def test_run_with_args(monkeypatch, run_command, first_app_built):
    """A static web app can be launched as a server; passthrough args will be
    ignored."""
    # Mock server creation
    mock_server_init = mock.MagicMock(spec_set=HTTPServer)
    monkeypatch.setattr(HTTPServer, "__init__", mock_server_init)

    # Mock the socket name returned by the server.
    socket = mock.MagicMock()
    socket.getsockname.return_value = ("127.0.0.1", "8080")
    LocalHTTPServer.socket = socket

    # Mock server execution, raising a user exit.
    mock_serve_forever = mock.MagicMock(side_effect=KeyboardInterrupt())
    monkeypatch.setattr(HTTPServer, "serve_forever", mock_serve_forever)

    # Mock shutdown
    mock_shutdown = mock.MagicMock()
    monkeypatch.setattr(HTTPServer, "shutdown", mock_shutdown)

    # Mock server close
    mock_server_close = mock.MagicMock()
    monkeypatch.setattr(HTTPServer, "server_close", mock_server_close)

    # Mock the webbrowser
    mock_open_new_tab = mock.MagicMock()
    monkeypatch.setattr(webbrowser, "open_new_tab", mock_open_new_tab)

    # Run the app
    run_command.run_app(
        first_app_built,
        test_mode=False,
        passthrough=["foo", "--bar"],
        host="localhost",
        port=8080,
        open_browser=True,
    )

    # The browser was opened
    mock_open_new_tab.assert_called_once_with("http://127.0.0.1:8080")

    # The server was started
    mock_serve_forever.assert_called_once_with()

    # The webserver was shutdown.
    mock_shutdown.assert_called_once_with()

    # The webserver was closed.
    mock_server_close.assert_called_once_with()


@pytest.mark.parametrize(
    "host, port, exception, message",
    [
        (
            "localhost",
            80,
            PermissionError(),
            r"Try using a port > 1023\.",
        ),
        (
            "localhost",
            8080,
            PermissionError(),
            r"Did you specify a valid host and port\?",
        ),
        (
            "999.999.999.999",
            8080,
            ErrnoError(errno.EADDRNOTAVAIL),
            r"999.999.999.999 is not a valid hostname.",
        ),
        (
            "999.999.999.999",
            8080,
            ErrnoError(errno.ENOSTR),
            r"999.999.999.999 is not a valid hostname.",
        ),
        (
            "localhost",
            8080,
            OSError("Unknown error"),
            r"Unknown error",
        ),
        (
            "localhost",
            99999,
            OverflowError(),
            r"Port must be in the range 0-65535.",
        ),
    ],
)
def test_cleanup_server_error(
    monkeypatch,
    run_command,
    first_app_built,
    host,
    port,
    exception,
    message,
):
    """If the server raises an error, it is cleaned up."""
    # Mock server creation, raising an error.
    mock_server_init = mock.MagicMock(side_effect=exception)
    monkeypatch.setattr(HTTPServer, "__init__", mock_server_init)

    # Mock server execution
    mock_serve_forever = mock.MagicMock()
    monkeypatch.setattr(HTTPServer, "serve_forever", mock_serve_forever)

    # Mock shutdown
    mock_shutdown = mock.MagicMock()
    monkeypatch.setattr(HTTPServer, "shutdown", mock_shutdown)

    # Mock server close
    mock_server_close = mock.MagicMock()
    monkeypatch.setattr(HTTPServer, "server_close", mock_server_close)

    # Mock the webbrowser
    mock_open_new_tab = mock.MagicMock()
    monkeypatch.setattr(webbrowser, "open_new_tab", mock_open_new_tab)

    # Run the app; an error is raised
    with pytest.raises(BriefcaseCommandError, match=message):
        run_command.run_app(
            first_app_built,
            test_mode=False,
            passthrough=[],
            host=host,
            port=port,
            open_browser=True,
        )

    # The browser was not opened
    mock_open_new_tab.assert_not_called()

    # The server was not started
    mock_serve_forever.assert_not_called()

    # The webserver was never started, so it wasn't shut down either.
    mock_shutdown.assert_not_called()
    mock_server_close.assert_not_called()


def test_cleanup_runtime_server_error(monkeypatch, run_command, first_app_built):
    """If the server raises an error at runtime, it is cleaned up."""
    # Mock server creation, raising an error due to an already used port.
    mock_server_init = mock.MagicMock()
    monkeypatch.setattr(HTTPServer, "__init__", mock_server_init)

    # Mock the socket name returned by the server.
    socket = mock.MagicMock()
    socket.getsockname.return_value = ("127.0.0.1", "8080")
    LocalHTTPServer.socket = socket

    # Mock server execution
    mock_serve_forever = mock.MagicMock(side_effect=ValueError())
    monkeypatch.setattr(HTTPServer, "serve_forever", mock_serve_forever)

    # Mock shutdown
    mock_shutdown = mock.MagicMock()
    monkeypatch.setattr(HTTPServer, "shutdown", mock_shutdown)

    # Mock server close
    mock_server_close = mock.MagicMock()
    monkeypatch.setattr(HTTPServer, "server_close", mock_server_close)

    # Mock the webbrowser
    mock_open_new_tab = mock.MagicMock()
    monkeypatch.setattr(webbrowser, "open_new_tab", mock_open_new_tab)

    # Run the app; it raises an error
    with pytest.raises(ValueError):
        run_command.run_app(
            first_app_built,
            test_mode=False,
            passthrough=[],
            host="localhost",
            port=8080,
            open_browser=True,
        )

    # The browser was opened
    mock_open_new_tab.assert_called_once_with("http://127.0.0.1:8080")

    # The server was started
    mock_serve_forever.assert_called_once_with()

    # The server crashed, so it won't need to be shut down
    mock_shutdown.assert_not_called()

    # The webserver was closed.
    mock_server_close.assert_called_once_with()


def test_run_without_browser(monkeypatch, run_command, first_app_built):
    """A static web app can be launched as a server."""
    # Mock server creation
    mock_server_init = mock.MagicMock()
    monkeypatch.setattr(HTTPServer, "__init__", mock_server_init)

    # Mock the socket name returned by the server.
    socket = mock.MagicMock()
    socket.getsockname.return_value = ("127.0.0.1", "8080")
    LocalHTTPServer.socket = socket

    # Mock server execution, raising a user exit.
    mock_serve_forever = mock.MagicMock(side_effect=KeyboardInterrupt())
    monkeypatch.setattr(HTTPServer, "serve_forever", mock_serve_forever)

    # Mock shutdown
    mock_shutdown = mock.MagicMock()
    monkeypatch.setattr(HTTPServer, "shutdown", mock_shutdown)

    # Mock server close
    mock_server_close = mock.MagicMock()
    monkeypatch.setattr(HTTPServer, "server_close", mock_server_close)

    # Mock the webbrowser
    mock_open_new_tab = mock.MagicMock()
    monkeypatch.setattr(webbrowser, "open_new_tab", mock_open_new_tab)

    # Run the app
    run_command.run_app(
        first_app_built,
        test_mode=False,
        passthrough=[],
        host="localhost",
        port=8080,
        open_browser=False,
    )

    # The browser was not opened
    mock_open_new_tab.assert_not_called()

    # The server was started
    mock_serve_forever.assert_called_once_with()

    # The webserver was shut down.
    mock_shutdown.assert_called_once_with()

    # The webserver was closed.
    mock_server_close.assert_called_once_with()


def test_run_autoselect_port(monkeypatch, run_command, first_app_built):
    """A static web app can be launched as a server."""
    # Mock server creation
    mock_server_init = mock.MagicMock()
    monkeypatch.setattr(HTTPServer, "__init__", mock_server_init)

    # Mock the socket name returned by the server.
    # This value has been auto-selected by the server.
    socket = mock.MagicMock()
    socket.getsockname.return_value = ("127.0.0.1", "12345")
    LocalHTTPServer.socket = socket

    # Mock server execution, raising a user exit.
    mock_serve_forever = mock.MagicMock(side_effect=KeyboardInterrupt())
    monkeypatch.setattr(HTTPServer, "serve_forever", mock_serve_forever)

    # Mock shutdown
    mock_shutdown = mock.MagicMock()
    monkeypatch.setattr(HTTPServer, "shutdown", mock_shutdown)

    # Mock server close
    mock_server_close = mock.MagicMock()
    monkeypatch.setattr(HTTPServer, "server_close", mock_server_close)

    # Mock the webbrowser
    mock_open_new_tab = mock.MagicMock()
    monkeypatch.setattr(webbrowser, "open_new_tab", mock_open_new_tab)

    # Run the app on an autoselected port
    run_command.run_app(
        first_app_built,
        test_mode=False,
        passthrough=[],
        host="localhost",
        port=0,
        open_browser=True,
    )

    # The browser was opened
    mock_open_new_tab.assert_called_once_with("http://127.0.0.1:12345")

    # The server was started
    mock_serve_forever.assert_called_once_with()

    # The webserver was shut down.
    mock_shutdown.assert_called_once_with()

    # The webserver was closed.
    mock_server_close.assert_called_once_with()


def test_served_paths(monkeypatch, tmp_path):
    """URLs are converted into paths in the project www folder."""
    # Mock server creation
    mock_server_init = mock.MagicMock(return_value=None)
    monkeypatch.setattr(SimpleHTTPRequestHandler, "__init__", mock_server_init)

    # Create a handler instance.
    request = mock.MagicMock()
    server = mock.MagicMock()
    handler = HTTPHandler(request, ("localhost", 8080), server)

    # We need some properties that are set on the handler instance
    # by the superclass; force set them here for test purposes.
    handler.server = server
    handler.server.base_path = tmp_path / "base_path"

    # Invoke this as a static method because we don't want to
    # instantiate a full server just to verify that URL rewriting works.
    assert handler.translate_path("/static/css/briefcase.css") == str(
        tmp_path / "base_path/static/css/briefcase.css"
    )


def test_cache_headers(monkeypatch, tmp_path):
    """Server sets no-cache headers."""
    # Mock server creation
    mock_server_init = mock.MagicMock(return_value=None)
    monkeypatch.setattr(SimpleHTTPRequestHandler, "__init__", mock_server_init)

    # Mock end_headers on the base class
    mock_end_headers = mock.MagicMock()
    monkeypatch.setattr(SimpleHTTPRequestHandler, "end_headers", mock_end_headers)

    # Create a handler instance.
    request = mock.MagicMock()
    server = mock.MagicMock()
    handler = HTTPHandler(request, ("localhost", 8080), server)

    # We need some properties that are set on the handler instance
    # by the superclass; force set them here for test purposes.
    handler.request_version = "HTTP/1.1"

    # Invoke end_headers()
    handler.end_headers()

    # end_headers was invoked on the base class...
    mock_end_headers.assert_called_once_with()

    # ...but the custom handler added cache control headers.
    assert handler._headers_buffer == [
        b"Cache-Control: no-cache, no-store, must-revalidate\r\n",
        b"Pragma: no-cache\r\n",
        b"Expires: 0\r\n",
    ]


def test_log_requests_to_logger(monkeypatch):
    """The request handler logs messages to the server's logger."""
    monkeypatch.setattr(
        SimpleHTTPRequestHandler, "handle", mock.Mock(return_value=None)
    )
    server = mock.MagicMock()
    handler = HTTPHandler(mock.MagicMock(), ("localhost", 8080), server)
    handler.log_date_time_string = mock.Mock(return_value="now")
    handler.log_message("hello\033")
    server.logger.info.assert_called_once_with("localhost - - [now] hello\\x1b")


def test_test_mode(run_command, first_app_built):
    """Test mode raises an error (at least for now)."""
    # Run the app
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Briefcase can't run web apps in test mode.",
    ):
        run_command.run_app(
            first_app_built,
            test_mode=True,
            passthrough=[],
            host="localhost",
            port=8080,
            open_browser=True,
        )



import platform
from subprocess import CalledProcessError

import pytest

from briefcase.exceptions import BriefcaseCommandError


@pytest.mark.parametrize(
    "host_os, host_arch",
    [
        ("Windows", "arm64"),
        ("Linux", "armv7l"),
    ],
)
def test_unsupported_abi(mock_tools, android_sdk, host_os, host_arch):
    """If the system has an unsupported ABI, raise an error."""
    # Mock the hardware and OS
    mock_tools.host_os = host_os
    mock_tools.host_arch = host_arch

    with pytest.raises(
        BriefcaseCommandError,
        match=f"The Android emulator does not currently support {host_os} {host_arch} hardware",
    ):
        android_sdk.verify_system_image("system-images;android-31;default;x86_64")


@pytest.mark.parametrize(
    "bad_image_name",
    [
        # Patently invalid
        "cheesecake",
        # Only the prefix
        "system-images",
        # Not enough parts
        "system-images;android-31;default",
        # Broadly good structure, but bad initial prefix
        "system-image;android-31;default;anything",
    ],
)
def test_invalid_system_image(mock_tools, android_sdk, bad_image_name):
    """If the system image name doesn't make sense, raise an error."""
    # Mock the host arch
    mock_tools.host_arch = "x86_64"

    # Verify a system image that doesn't match the host architecture
    with pytest.raises(
        BriefcaseCommandError,
        match=rf"{bad_image_name!r} is not a valid system image name.",
    ):
        android_sdk.verify_system_image(bad_image_name)


def test_incompatible_abi(mock_tools, android_sdk, capsys):
    """If the system image doesn't match the emulator ABI, warn the user, but
    continue."""
    # Mock the host arch
    mock_tools.host_arch = "AMD64" if platform.system() == "Windows" else "x86_64"

    # Verify a system image that doesn't match the host architecture
    android_sdk.verify_system_image("system-images;android-31;default;anything")

    # A warning message was output
    assert "WARNING: Unexpected emulator ABI" in capsys.readouterr().out

    # The system image was installed.
    mock_tools.subprocess.run.assert_called_once_with(
        [
            android_sdk.sdkmanager_path,
            "system-images;android-31;default;anything",
        ],
        env=android_sdk.env,
        check=True,
        stream_output=False,
    )


def test_existing_system_image(mock_tools, android_sdk):
    """If the system image already exists, don't attempt to download it again."""
    # Mock the host arch
    mock_tools.host_arch = "AMD64" if platform.system() == "Windows" else "x86_64"

    # Mock the existence of a system image
    (android_sdk.root_path / "system-images/android-31/default/x86_64").mkdir(
        parents=True
    )

    # Verify the system image that we already have
    android_sdk.verify_system_image("system-images;android-31;default;x86_64")

    # sdkmanager was *not* called.
    mock_tools.subprocess.run.assert_not_called()


def test_new_system_image(mock_tools, android_sdk):
    """If the system image doesn't exist locally, it will be installed."""
    # Mock the host arch
    mock_tools.host_arch = "AMD64" if platform.system() == "Windows" else "x86_64"

    # Verify the system image, triggering a download
    android_sdk.verify_system_image("system-images;android-31;default;x86_64")

    # The system image was installed.
    mock_tools.subprocess.run.assert_called_once_with(
        [
            android_sdk.sdkmanager_path,
            "system-images;android-31;default;x86_64",
        ],
        env=android_sdk.env,
        check=True,
        stream_output=False,
    )


def test_problem_downloading_system_image(mock_tools, android_sdk):
    """If there is a failure downloading the system image, an error is raised."""
    # Mock the host arch
    mock_tools.host_arch = "AMD64" if platform.system() == "Windows" else "x86_64"

    # Mock a failure condition on subprocess.run
    mock_tools.subprocess.run.side_effect = CalledProcessError(
        returncode=1,
        cmd="sdkmanager",
    )

    # Attempt to verify the system image
    with pytest.raises(
        BriefcaseCommandError,
        match="Error while installing the 'system-images;android-31;default;x86_64' Android system image.",
    ):
        android_sdk.verify_system_image("system-images;android-31;default;x86_64")

    # An attempt to install the system image was made
    mock_tools.subprocess.run.assert_called_once_with(
        [
            android_sdk.sdkmanager_path,
            "system-images;android-31;default;x86_64",
        ],
        env=android_sdk.env,
        check=True,
        stream_output=False,
    )



import os
from pathlib import Path

import pytest

from briefcase.platforms.macOS import SigningIdentity

from ....utils import create_file, create_plist_file


@pytest.fixture
def sekrit_identity():
    return SigningIdentity(id="CAFEBEEF", name="Sekrit identity (DEADBEEF)")


@pytest.fixture
def sekrit_installer_identity():
    return SigningIdentity(id="CAFEFACE", name="Sekrit Installer identity (DEADBEEF)")


@pytest.fixture
def adhoc_identity():
    return SigningIdentity()


@pytest.fixture
def first_app_templated(first_app_config, tmp_path):
    app_path = tmp_path / "base_path/build/first-app/macos/app/First App.app"

    # Create the briefcase.toml file
    create_file(
        tmp_path / "base_path/build/first-app/macos/app/briefcase.toml",
        """
[paths]
app_packages_path="First App.app/Contents/Resources/app_packages"
support_path="First App.app/Contents/Resources/support"
info_plist_path="First App.app/Contents/Info.plist"
entitlements_path="Entitlements.plist"
""",
    )

    # Create the plist file for the app
    create_plist_file(
        app_path / "Contents/Info.plist",
        {
            "MainModule": "first_app",
        },
    )

    # Create the entitlements file for the app
    create_plist_file(
        tmp_path / "base_path/build/first-app/macos/app/Entitlements.plist",
        {
            "com.apple.security.cs.allow-unsigned-executable-memory": True,
            "com.apple.security.cs.disable-library-validation": True,
        },
    )

    # Create some folders that need to exist.
    (app_path / "Contents/Resources/app_packages").mkdir(parents=True)
    (app_path / "Contents/Frameworks").mkdir(parents=True)

    # Create an installer Distribution.xml
    create_file(
        tmp_path / "base_path/build/first-app/macos/app/installer/Distribution.xml",
        """<?xml?>\n<installer-script></installer-script>""",
    )

    # Select dmg packaging by default
    first_app_config.packaging_format = "dmg"

    return first_app_config


@pytest.fixture
def first_app_with_binaries(first_app_templated, first_app_config, tmp_path):
    app_path = tmp_path / "base_path/build/first-app/macos/app/First App.app"

    # Create the stub binary
    create_file(app_path / "Contents/MacOS/First App", "Stub binary")

    # Create some libraries that need to be signed.
    lib_path = app_path / "Contents/Resources/app_packages"
    frameworks_path = app_path / "Contents/Frameworks"

    for lib in [
        "first_so.so",
        Path("subfolder/second_so.so"),
        "first_dylib.dylib",
        Path("subfolder/second_dylib.dylib"),
        "other_binary",
    ]:
        (lib_path / lib).parent.mkdir(parents=True, exist_ok=True)
        with (lib_path / lib).open("wb") as f:
            f.write(b"\xCA\xFE\xBA\xBEBinary content here")

    # Mach-O file that is executable, with an odd extension
    with (lib_path / "special.binary").open("wb") as f:
        f.write(b"\xCA\xFE\xBA\xBEBinary content here")
    os.chmod(lib_path / "special.binary", 0o755)

    # An embedded app
    (lib_path / "Extras.app/Contents/MacOS").mkdir(parents=True, exist_ok=True)
    with (lib_path / "Extras.app/Contents/MacOS/Extras").open("wb") as f:
        f.write(b"\xCA\xFE\xBA\xBEBinary content here")

    # An embedded framework
    (frameworks_path / "Extras.framework/Resources").mkdir(parents=True, exist_ok=True)
    with (frameworks_path / "Extras.framework/Resources/extras.dylib").open("wb") as f:
        f.write(b"\xCA\xFE\xBA\xBEBinary content here")

    # Make sure there are some files in the bundle that *don't* need to be signed...
    with (lib_path / "first.other").open("w", encoding="utf-8") as f:
        f.write("other")
    with (lib_path / "second.other").open("w", encoding="utf-8") as f:
        f.write("other")

    # A file that has a Mach-O header, but isn't executable
    with (lib_path / "unknown.binary").open("wb") as f:
        f.write(b"\xCA\xFE\xBA\xBEother")

    return first_app_config



from __future__ import annotations

import argparse
import importlib
import importlib.metadata
import inspect
import os
import platform
import subprocess
import sys
from abc import ABC, abstractmethod
from argparse import RawDescriptionHelpFormatter
from pathlib import Path
from typing import Any

from cookiecutter import exceptions as cookiecutter_exceptions
from cookiecutter.repository import is_repo_url
from packaging.version import Version
from platformdirs import PlatformDirs

if sys.version_info >= (3, 11):  # pragma: no-cover-if-lt-py311
    import tomllib
else:  # pragma: no-cover-if-gte-py311
    import tomli as tomllib

import briefcase
from briefcase import __version__
from briefcase.config import AppConfig, GlobalConfig, parse_config
from briefcase.console import MAX_TEXT_WIDTH, Console, Log
from briefcase.exceptions import (
    BriefcaseCommandError,
    BriefcaseConfigError,
    InvalidTemplateRepository,
    MissingAppMetadata,
    NetworkFailure,
    TemplateUnsupportedVersion,
    UnsupportedHostError,
)
from briefcase.integrations.base import ToolCache
from briefcase.integrations.file import File
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms import get_output_formats, get_platforms


def create_config(klass, config, msg):
    try:
        return klass(**config)
    except TypeError as e:
        # Inspect the GlobalConfig constructor to find which
        # parameters are required and don't have a default value.
        required_args = {
            name
            for name, param in inspect.signature(klass.__init__).parameters.items()
            if param.default == inspect._empty and name not in {"self", "kwargs"}
        }
        missing_args = required_args - config.keys()
        missing = ", ".join(f"'{arg}'" for arg in sorted(missing_args))
        raise BriefcaseConfigError(f"{msg} is incomplete (missing {missing})") from e


def full_options(state, options):
    """Merge command state with keyword arguments.

    Command state takes precedence over any keyword argument.

    :param state: The current command state. Can be ``None``.
    :param options: The base options.
    :returns: A dictionary containing all of ``options``, with any values
        provided in ``state`` overriding the base ``options`` values.
    """
    if state is not None:
        full = options.copy()
        full.update(state)
    else:
        full = options

    return full


def split_passthrough(args):
    try:
        pos = args.index("--")
    except ValueError:
        return args, []
    else:
        return args[:pos], args[pos + 1 :]


def parse_config_overrides(config_overrides: list[str] | None) -> dict[str, Any]:
    """Parse command line -C/--config option overrides.

    :param config_overrides: The values passed in as configuration overrides. Each value
        *should* be a "key=<valid TOML>" string.
    :returns: A dictionary of app configuration keys to override and their new values.
    :raises BriefcaseCommandError: if any of the values can't be parsed as valid TOML.
    """
    overrides = {}
    if config_overrides:
        for override in config_overrides:
            try:
                # Do initial checks of the key being overridden.
                # These catch cases that would be valid TOML, but would result
                # in invalid app configurations.
                key, _ = override.split("=", 1)
                if "." in key:
                    raise BriefcaseConfigError(
                        "Can't override multi-level configuration keys."
                    )
                elif key == "app_name":
                    raise BriefcaseConfigError("The app name cannot be overridden.")

                # Now actually parse the value
                overrides.update(tomllib.loads(override))
            except ValueError as e:
                raise BriefcaseConfigError(
                    f"Unable to parse configuration override {override}"
                ) from e
    return overrides


class BaseCommand(ABC):
    cmd_line = "briefcase {command} {platform} {output_format}"
    supported_host_os = {"Darwin", "Linux", "Windows"}
    supported_host_os_reason = f"This command is not supported on {platform.system()}."

    # defined by platform-specific subclasses
    command: str
    description: str
    platform: str
    output_format: str
    # supports passing extra command line arguments to subprocess
    allows_passthrough = False
    # if specified for a platform, then any template for that platform must declare
    # compatibility with that version epoch. An epoch begins when a breaking change is
    # introduced for a platform such that older versions of a template are incompatible
    platform_target_version: str | None = None

    def __init__(
        self,
        logger: Log,
        console: Console,
        tools: ToolCache = None,
        apps: dict = None,
        base_path: Path = None,
        data_path: Path = None,
        is_clone: bool = False,
    ):
        """Base for all Commands.

        :param logger: Logger for console and logfile.
        :param console: Facilitates console interaction and input solicitation.
        :param tools: Cache of tools populated by Commands as they are required.
        :param apps: Dictionary of project's Apps keyed by app name.
        :param base_path: Base directory for Briefcase project.
        :param data_path: Base directory for Briefcase tools, support packages, etc.
        :param is_clone: Flag that Command was triggered by the user's requested
            Command; for instance, RunCommand can invoke UpdateCommand and/or
            BuildCommand.
        """
        if base_path is None:
            self.base_path = Path.cwd()
        else:
            self.base_path = base_path
        self.data_path = self.validate_data_path(data_path)
        self.apps = {} if apps is None else apps
        self.is_clone = is_clone

        self.tools = tools or ToolCache(
            logger=logger,
            console=console,
            base_path=self.data_path / "tools",
        )

        # Immediately add tools that must be always available
        Subprocess.verify(tools=self.tools)
        File.verify(tools=self.tools)

        if not is_clone:
            self.validate_locale()

        self.global_config = None
        self._briefcase_toml: dict[AppConfig, dict[str, ...]] = {}

    @property
    def logger(self):
        return self.tools.logger

    @property
    def input(self):
        return self.tools.input

    def validate_data_path(self, data_path):
        """Validate provided data path or determine OS-specific path.

        If a data path is provided during construction, use it. This usually indicates
        we're under test conditions. If there's no data path provided, look for a
        BRIEFCASE_HOME environment variable. If that isn't defined, use a platform-
        specific default data path.
        """
        if data_path is None:
            try:
                briefcase_home = os.environ["BRIEFCASE_HOME"]
                data_path = Path(briefcase_home).resolve()
                # Path("") converts to ".", so check for that edge case.
                if briefcase_home == "" or not data_path.is_dir():
                    raise BriefcaseCommandError(
                        "The path specified by BRIEFCASE_HOME does not exist."
                    )
            except KeyError:
                if platform.system() == "Darwin":  # pragma: no-cover-if-not-macos
                    # macOS uses a bundle name, rather than just the app name
                    app_name = "org.beeware.briefcase"
                else:  # pragma: no-cover-if-is-macos
                    app_name = "briefcase"

                data_path = PlatformDirs(
                    appname=app_name,
                    appauthor="BeeWare",
                ).user_cache_path

        data_path = os.fsdecode(data_path)

        if " " in data_path:
            raise BriefcaseCommandError(
                f"""
The location Briefcase will use to store tools and support files:

    {data_path}

contains spaces. This will cause problems with some tools, preventing
you from building and packaging applications.

You can set the environment variable BRIEFCASE_HOME to specify
a custom location for Briefcase's tools.

"""
            )

        if not os.path.exists(data_path):
            try:
                # The Windows Store version of Python can redirect filesystem
                # interactions within %LOCALAPPDATA% to a sandboxed location.
                # To bypass this, the Briefcase cache directory creation is
                # performed via ``cmd.exe`` in a different process. Once this
                # directory exists in the "real" %LOCALAPPDATA%, Windows will
                # allow normal interactions without attempting to sandbox them.
                if platform.system() == "Windows":  # pragma: no-cover-if-not-windows
                    subprocess.run(
                        ["mkdir", data_path],
                        shell=True,
                        check=True,
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                    )
                else:  # pragma: no-cover-if-is-windows
                    os.makedirs(data_path, exist_ok=True)
            except (subprocess.CalledProcessError, OSError):
                raise BriefcaseCommandError(
                    f"""
Failed to create the Briefcase directory to store tools and support files:

    {data_path}

You can set the environment variable BRIEFCASE_HOME to specify
a custom location for Briefcase's tools.

"""
                )

        return Path(data_path)

    def validate_locale(self):
        """Validate the system's locale is compatible."""
        if self.tools.host_os == "Linux" and self.tools.system_encoding != "UTF-8":
            self.logger.warning(
                f"""
*************************************************************************
** WARNING: Default system encoding is not supported                   **
*************************************************************************

    Briefcase and the third-party tools it uses only support UTF-8.

    The detected default system encoding is {self.tools.system_encoding}.

    Briefcase will proceed but some console output could be corrupted and
    created files or artefacts may contain corrupted text.

    Update your system's encoding to UTF-8 to avoid issues.

*************************************************************************
"""
            )

    def _command_factory(self, command_name: str):
        """Command factory for the current platform and format.

        :param command_name: name of Command (e.g. 'create', 'build', 'run', etc.)
        :return: instantiated Command
        """
        format_module = importlib.import_module(self.__module__)
        command = getattr(format_module, command_name)(
            base_path=self.base_path,
            apps=self.apps,
            logger=self.logger,
            console=self.input,
            tools=self.tools,
            is_clone=True,
        )
        command.clone_options(self)
        return command

    @property
    def create_command(self):
        """Create Command factory for the same platform and format."""
        return self._command_factory("create")

    @property
    def update_command(self):
        """Update Command factory for the same platform and format."""
        return self._command_factory("update")

    @property
    def build_command(self):
        """Build Command factory for the same platform and format."""
        return self._command_factory("build")

    @property
    def run_command(self):
        """Run Command factory for the same platform and format."""
        return self._command_factory("run")

    @property
    def package_command(self):
        """Package Command factory for the same platform and format."""
        return self._command_factory("package")

    @property
    def publish_command(self):
        """Publish Command factory for the same platform and format."""
        return self._command_factory("publish")

    def template_cache_path(self, template) -> Path:
        """The path where Briefcase keeps template checkouts.

        :param template: The URL for the template that will be cached locally.
        """
        template = template.rstrip("/")
        tail = template.split("/")[-1]
        cache_name = tail.rsplit(".git")[0]
        return self.data_path / "templates" / cache_name

    def build_path(self, app) -> Path:
        """The path in which all platform artefacts for the app will be built.

        :param app: The app config
        """
        return self.base_path / "build" / app.app_name / self.platform.lower()

    @property
    def dist_path(self) -> Path:
        """The path for all applications for this command's platform."""
        return self.base_path / "dist"

    def bundle_path(self, app) -> Path:
        """The path to the bundle for the app in the output format.

        The bundle is the template-generated source form of the app. The path will
        usually be a directory, the existence of which is indicative that the template
        has been rolled out for an app. The leaf of this path is the base of the content
        generated from template.

        :param app: The app config
        """
        return self.build_path(app) / self.output_format.lower()

    @abstractmethod
    def binary_path(self, app) -> Path:
        """The path to the executable artefact for the app in the output format.

        This may be a binary file produced by compilation; however, if the output format
        doesn't require compilation, it may be the same as the bundle path (assuming the
        bundle path is inherently "executable"), or a path that reasonably represents
        the thing that can be executed.

        :param app: The app config
        """

    def binary_executable_path(self, app) -> Path:
        """The path to the actual binary object for the app in the output format.

        For most platforms, this will be the same as the binary path. However, for
        platforms that use an "executable bundle" (e.g., macOS), this will be actual
        binary that is embedded in the bundle.

        :param app: The app config
        """
        return self.binary_path(app)

    def unbuilt_executable_path(self, app) -> Path:
        """The path to the unbuilt form of the binary object for the app.

        The pre-built stub binary may need to undergo some manipulation before it can be
        used; to mark that this manipulation is required, the "unbuilt" binary has a
        "raw" name that doesn't involve any app details. The build step moves the binary
        to the final name.

        :param app: The app config
        """
        return self.binary_executable_path(app).parent / (
            "Stub" + self.binary_executable_path(app).suffix
        )

    def briefcase_toml(self, app: AppConfig) -> dict[str, ...]:
        """Load the ``briefcase.toml`` file provided by the app template.

        :param app: The config object for the app
        :return: The contents of ``briefcase.toml``
        """
        try:
            return self._briefcase_toml[app]
        except KeyError:
            try:
                with (self.bundle_path(app) / "briefcase.toml").open("rb") as f:
                    self._briefcase_toml[app] = tomllib.load(f)
            except OSError as e:
                raise MissingAppMetadata(self.bundle_path(app)) from e
            else:
                return self._briefcase_toml[app]

    def path_index(self, app: AppConfig, path_name: str) -> str | dict | list:
        """Return a path from the path index provided by the app template.

        Raises KeyError if ``path_name`` is not defined in the index.

        :param app: The config object for the app
        :param path_name: Name of the filepath to retrieve
        :return: filepath for requested path
        """
        return self.briefcase_toml(app)["paths"][path_name]

    def template_target_version(self, app: AppConfig) -> str | None:
        """The target version of Briefcase for the app from ``briefcase.toml``.

        This value represents a version epoch specific to the platform. An epoch begins
        when a breaking change is introduced. Therefore, this value would remain the
        version of Briefcase that introduced a breaking change for a template until
        another such change requires a new epoch.

        :param app: The config object for the app
        :return: target version or None if one isn't specified
        """
        try:
            return self.briefcase_toml(app)["briefcase"]["target_version"]
        except KeyError:
            return None

    def stub_binary_revision(self, app: AppConfig) -> str:
        """Obtain the stub binary revision that the template requires.

        :param app: The config object for the app
        :return: The stub binary revision required by the template.
        """
        return self.path_index(app, "stub_binary_revision")

    def support_path(self, app: AppConfig) -> Path:
        """Obtain the path into which the support package should be unpacked.

        :param app: The config object for the app
        :return: The full path where the support package should be unpacked.
        """
        return self.bundle_path(app) / self.path_index(app, "support_path")

    def support_revision(self, app: AppConfig) -> str:
        """Obtain the support package revision that the template requires.

        :param app: The config object for the app
        :return: The support revision required by the template.
        """
        return self.path_index(app, "support_revision")

    def cleanup_paths(self, app: AppConfig) -> list[str]:
        """Obtain the paths generated by the app template that should be cleaned up
        prior to release.

        :param app: The config object for the app
        :return: The list of path globs inside the app template that should be cleaned
            up.
        """
        return self.path_index(app, "cleanup_paths")

    def app_requirements_path(self, app: AppConfig) -> Path:
        """Obtain the path into which a requirements.txt file should be written.

        :param app: The config object for the app
        :return: The full path where the requirements.txt file should be written
        """
        return self.bundle_path(app) / self.path_index(app, "app_requirements_path")

    def app_packages_path(self, app: AppConfig) -> Path:
        """Obtain the path into which requirements should be installed.

        :param app: The config object for the app
        :return: The full path where application requirements should be installed.
        """
        return self.bundle_path(app) / self.path_index(app, "app_packages_path")

    def app_path(self, app: AppConfig) -> Path:
        """Obtain the path into which the application should be installed.

        :param app: The config object for the app
        :return: The full path where application code should be installed.
        """
        return self.bundle_path(app) / self.path_index(app, "app_path")

    def app_module_path(self, app: AppConfig) -> Path:
        """Find the path for the application module for an app.

        :param app: The config object for the app
        :returns: The Path to the dist-info folder.
        """
        app_home = [
            path.split("/")
            for path in app.sources
            if path.rsplit("/", 1)[-1] == app.module_name
        ]

        if len(app_home) == 0:
            raise BriefcaseCommandError(
                f"Unable to find code for application {app.app_name!r}"
            )
        elif len(app_home) == 1:
            path = Path(str(self.base_path), *app_home[0])
        else:
            raise BriefcaseCommandError(
                f"Multiple paths in sources found for application {app.app_name!r}"
            )

        return path

    @property
    def briefcase_required_python_version(self):
        """The major.minor of the minimum Python version required by Briefcase itself.

        This is extracted from packaging metadata.
        """
        # Native format is ">=3.8"
        return tuple(
            int(v)
            for v in importlib.metadata.metadata("briefcase")["Requires-Python"]
            .split("=")[1]
            .strip()
            .split(".")
        )

    @property
    def python_version_tag(self):
        """The major.minor of the Python version in use, as a string.

        This is used as a repository label/tag to identify the appropriate templates,
        etc. to use.
        """
        return (
            f"{self.tools.sys.version_info.major}.{self.tools.sys.version_info.minor}"
        )

    def verify_host(self):
        """Verify the host OS is supported by the Command."""
        if self.tools.host_os not in self.supported_host_os:
            raise UnsupportedHostError(self.supported_host_os_reason)

    def verify_tools(self):
        """Verify that the tools needed to run this Command exist.

        Raises MissingToolException if a required system tool is missing.
        """

    def finalize_app_config(self, app: AppConfig):
        """Finalize the application config.

        Some app configurations (notably, Linux system packages like .deb) have
        configurations that are deeper than other platforms, because they need to
        include components that are dependent on command-line arguments. They may also
        require the existence of system tools to complete configuration.

        The final app configuration merges those "deep" properties into the app
        configuration, and performs any other app-specific platform configuration and
        verification that is required as a result of command-line arguments.

        :param app: The app configuration to finalize.
        """

    def finalize(self, app: AppConfig | None = None):
        """Finalize Briefcase configuration.

        This will:

        1. Ensure that the host has been verified
        2. Ensure that the platform tools have been verified
        3. Ensure that app configurations have been finalized.

        App finalization will only occur once per invocation.

        :param app: If provided, the specific app configuration
            to finalize. By default, all apps will be finalized.
        """
        self.verify_host()
        self.verify_tools()

        if app is None:
            for app in self.apps.values():
                if hasattr(app, "__draft__"):
                    self.finalize_app_config(app)
                    delattr(app, "__draft__")
        else:
            if hasattr(app, "__draft__"):
                self.finalize_app_config(app)
                delattr(app, "__draft__")

    def verify_app(self, app: AppConfig):
        """Verify the app is compatible and the app tools are available.

        This is the last step of verification for a Command before running the Command's
        business logic. It runs _after_ pre-requisite Commands have been verified and
        run.

        :param app: app configuration
        """
        self.verify_app_template(app)
        self.verify_app_tools(app)

    def verify_app_tools(self, app: AppConfig):
        """Verify that tools needed to run the command for this app exist."""

    def verify_app_template(self, app: AppConfig):
        """Verify the template targets the same Briefcase version as the Command.

        :param app: app configuration
        """

        # Skip this check if the template isn't rolled out
        # or if the command doesn't support templates
        try:
            template_target_version = self.template_target_version(app)
        except (MissingAppMetadata, NotImplementedError):
            return

        if self.platform_target_version != template_target_version:
            raise BriefcaseCommandError(
                f"""\
The app template used to generate this app is not compatible with this version
of Briefcase.

If the app was generated with an earlier version of Briefcase using the default
Briefcase template, you can run:

     $ briefcase create {self.platform} {self.output_format}

to re-generate your app with a compatible version of the template.

If you are using a custom template, you'll need to update the template to correct
any compatibility problems, and then add the compatibility declaration.
"""
            )

    def parse_options(self, extra):
        """Parse the command line arguments for the Command.

        After the initial ArgumentParser runs to choose the Command for the selected
        platform and format, a new ArgumentParser is created here to parse the remaining
        command line arguments specific to the Command. Additionally, the default
        options for disabling input, log verbosity, and log saving are parsed out and
        saved to the Command.

        :param extra: the remaining command line arguments after the initial
            ArgumentParser runs over the command line.
        :return: dictionary of parsed arguments for Command, and a dictionary of parsed
            configuration overrides.
        """
        default_format = getattr(
            get_platforms().get(self.platform), "DEFAULT_OUTPUT_FORMAT", None
        )
        # only show supported formats for Commands that support formats
        if default_format is not None and self.command not in {"new", "dev", "upgrade"}:
            formats = list(get_output_formats(self.platform).keys())
            formats[formats.index(default_format)] = f"{default_format} (default)"
            supported_formats_helptext = (
                f"Supported formats:\n  {', '.join(sorted(formats, key=str.lower))}"
            )
        else:
            supported_formats_helptext = ""

        parser = argparse.ArgumentParser(
            prog=self.cmd_line.format(
                command=self.command,
                platform=self.platform,
                output_format=self.output_format,
            ),
            description=self.input.textwrap(
                f"{self.description}\n\n{supported_formats_helptext}"
            ),
            formatter_class=(
                lambda prog: RawDescriptionHelpFormatter(prog, width=MAX_TEXT_WIDTH)
            ),
        )

        self.add_default_options(parser)
        self.add_options(parser)

        # If the command allows passthrough arguments, add an option for the help,
        # then process the argument list to strip out the passthrough args.
        if self.allows_passthrough:
            parser.add_argument(
                "--",
                dest="passthrough",
                metavar="ARGS ...",
                required=False,
                help="Arguments to pass to the app",
            )
            args, passthrough = split_passthrough(extra)
        else:
            args = extra

        # Parse the full set of command line options from the content
        # remaining after the basic command/platform/output format
        # has been extracted.
        options = vars(parser.parse_args(args))

        if self.allows_passthrough:
            options["passthrough"] = passthrough

        # Extract the base default options onto the command
        self.input.enabled = options.pop("input_enabled")
        self.logger.verbosity = options.pop("verbosity")
        self.logger.save_log = options.pop("save_log")

        # Parse the configuration overrides
        overrides = parse_config_overrides(options.pop("config_overrides"))

        return options, overrides

    def clone_options(self, command):
        """Clone options from one command to this one.

        :param command: The command whose options are to be cloned
        """

    def add_default_options(self, parser):
        """Add the default options that exist on *all* commands.

        :param parser: a stub argparse parser for the command.
        """
        parser.add_argument(
            "-C",
            "--config",
            dest="config_overrides",
            action="append",
            metavar="KEY=VALUE",
            help="Override the value of the app configuration item KEY with VALUE",
        )
        parser.add_argument(
            "-v",
            "--verbosity",
            action="count",
            default=0,
            help="Enable verbose logging. Use -vv and -vvv to increase logging verbosity",
        )
        parser.add_argument("-V", "--version", action="version", version=__version__)
        parser.add_argument(
            "--no-input",
            action="store_false",
            default=True,
            dest="input_enabled",
            help=(
                "Don't ask for user input. If any action would be destructive, "
                "an error will be raised; otherwise, default answers will be "
                "assumed"
            ),
        )
        parser.add_argument(
            "--log",
            action="store_true",
            dest="save_log",
            help="Save a detailed log to file. By default, this log file is only created for critical errors",
        )

    def _add_update_options(
        self,
        parser,
        context_label="",
        update=True,
    ):
        """Internal utility method for adding common update options.

        :param parser: The parser to which options should be added.
        :param context_label: Label text that will be added to the end of the help text
            to describe when the update will be applied (e.g., "before building")
        :param update: Should the --update and --no-update options be exposed?
        """
        if update:
            parser.add_argument(
                "-u",
                "--update",
                action="store_true",
                help=f"Update the app{context_label}",
            )

        parser.add_argument(
            "-r",
            "--update-requirements",
            action="store_true",
            help=f"Update requirements for the app{context_label}",
        )

        parser.add_argument(
            "--update-support",
            action="store_true",
            help=f"Update support package for the app{context_label}",
        )

        parser.add_argument(
            "--update-stub",
            action="store_true",
            help=f"Update stub binary for the app{context_label}",
        )

        parser.add_argument(
            "--update-resources",
            action="store_true",
            help=f"Update app resources (icons, splash screens, etc){context_label}",
        )

        if update:
            parser.add_argument(
                "--no-update",
                action="store_true",
                help=f"Prevent any automated update{context_label}",
            )

    def _add_test_options(self, parser, context_label):
        """Internal utility method for adding common test-related options.

        :param parser: The parser to which options should be added.
        :param context_label: Label text for commands; the capitalized action being
            performed (e.g., "Build", "Run",...)
        """
        parser.add_argument(
            "--test",
            dest="test_mode",
            action="store_true",
            help=f"{context_label} the app in test mode",
        )

    def add_options(self, parser):
        """Add any options that this command needs to parse from the command line.

        :param parser: a stub argparse parser for the command.
        """

    def parse_config(self, filename, overrides):
        try:
            with open(filename, "rb") as config_file:
                # Parse the content of the pyproject.toml file, extracting
                # any platform and output format configuration for each app,
                # creating a single set of configuration options.
                global_config, app_configs = parse_config(
                    config_file,
                    platform=self.platform,
                    output_format=self.output_format,
                    logger=self.logger,
                )

                # Create the global config
                global_config.update(overrides)
                self.global_config = create_config(
                    klass=GlobalConfig,
                    config=global_config,
                    msg="Global configuration",
                )

                for app_name, app_config in app_configs.items():
                    # Construct an AppConfig object with the final set of
                    # configuration options for the app.
                    app_config.update(overrides)
                    self.apps[app_name] = create_config(
                        klass=AppConfig,
                        config=app_config,
                        msg=f"Configuration for '{app_name}'",
                    )

        except OSError as e:
            raise BriefcaseConfigError(
                f"""\
Configuration file not found.

Did you run Briefcase in a project directory that contains {filename.name!r}?"""
            ) from e

    def update_cookiecutter_cache(self, template: str, branch="master"):
        """Ensure that we have a current checkout of a template path.

        If the path is a local path, use the path as is.

        If the path is a URL, look for a local cache; if one exists, update it,
        including checking out the required branch.

        :param template: The template URL or path.
        :param branch: The template branch to use. Default: ``master``
        :return: The path to the cached template. This may be the originally
            provided path if the template was a file path.
        """
        if is_repo_url(template):
            # The app template is a repository URL.
            #
            # Look for a Briefcase cache of the template.
            cached_template = self.template_cache_path(template)

            if cached_template.exists():
                # There is a pre-existing cache of the template. Attempt to update it;
                # if the fetch fails, use the existing state of the cache. Any other
                # failure is surfaced to the user.
                try:
                    repo = self.tools.git.Repo(cached_template)
                except self.tools.git.exc.GitError:
                    # Template repository is in a weird state. Delete it
                    self.logger.warning(
                        "Template cache is in a weird state. Getting a clean clone."
                    )
                    self.tools.shutil.rmtree(cached_template)

            if not cached_template.exists():
                # There's no pre-existing template. It's either the first time seeing
                # the template, or the template was in a weird state. Perform a blobless
                # clone.
                try:
                    self.logger.info(f"Cloning template {template!r}...")
                    cached_template.mkdir(exist_ok=True, parents=True)
                    repo = self.tools.git.Repo.clone_from(
                        url=template,
                        to_path=cached_template,
                        filter=["blob:none"],
                        no_checkout=True,
                    )
                except KeyboardInterrupt:
                    # The user has aborted the initial clone. Git is fairly resilient to
                    # being interrupted, but if the *initial* clone fails, it's very
                    # hard to recover. To avoid problems on the next run, remove the
                    # partial repo clone.
                    if cached_template.exists():
                        self.tools.shutil.rmtree(cached_template)
                    raise
                except self.tools.git.exc.GitError as e:
                    # The clone failed; to make sure the repo is in a clean state, clean up
                    # any partial remnants of this initial clone.
                    # If we're getting a GitError, we know the directory must exist.
                    self.tools.shutil.rmtree(cached_template)
                    raise BriefcaseCommandError(
                        f"Unable to clone repository {template!r}.\n"
                        "\n"
                        "This may be because your computer is offline, or "
                        "because the repository URL is incorrect."
                    ) from e

            try:
                # Raises ValueError if "origin" isn't a valid remote
                remote = repo.remote(name="origin")
                # Ensure the existing repo's origin URL points to the location
                # being requested. A difference can occur, for instance, if a
                # fork of the template is used.
                remote.set_url(new_url=template, old_url=remote.url)
                try:
                    # Attempt to update the repository
                    remote.fetch()
                except self.tools.git.exc.GitCommandError as e:
                    # We are offline, or otherwise unable to contact the origin git
                    # repo. It's OK to continue; but capture the error in the log and
                    # warn the user that the template may be stale.
                    self.logger.debug(str(e))
                    self.logger.warning(
                        """
*************************************************************************
** WARNING: Unable to update template                                  **
*************************************************************************

    Briefcase is unable the update the application template. This
    may be because your computer is currently offline. Briefcase will
    use existing template without updating.

*************************************************************************
"""
                    )

                try:
                    # Check out the branch for the required version tag.
                    head = remote.refs[branch]

                    self.logger.info(
                        f"Using existing template (sha {head.commit.hexsha}, "
                        f"updated {head.commit.committed_datetime.strftime('%c')})"
                    )
                    head.checkout()
                except IndexError as e:
                    # No branch exists for the requested version.
                    raise TemplateUnsupportedVersion(branch) from e
            except (ValueError, self.tools.git.exc.GitError) as e:
                raise BriefcaseCommandError(
                    "Unable to check out template branch.\n"
                    "\n"
                    "This may be because your computer is offline, or because the template repository\n"
                    "is in a weird state. If you have a stable network connection, try deleting:\n"
                    "\n"
                    f"    {cached_template}\n"
                    "\n"
                    "and retrying your command."
                ) from e
        else:
            # If this isn't a repository URL, treat it as a local directory
            cached_template = template

        return cached_template

    def _generate_template(self, template, branch, output_path, extra_context):
        """Ensure the named template is up-to-date for the given branch, and roll out
        that template.

        :param template: The template URL or path to generate
        :param branch: The branch of the template to use
        :param output_path: The filesystem path where the template will be generated.
        :param extra_context: Extra context to pass to the cookiecutter template
        """
        # Make sure we have an updated cookiecutter template,
        # checked out to the right branch
        cached_template = self.update_cookiecutter_cache(
            template=template,
            branch=branch,
        )

        self.logger.configure_stdlib_logging("cookiecutter")

        try:
            # Unroll the template.
            self.tools.cookiecutter(
                str(cached_template),
                no_input=True,
                output_dir=str(output_path),
                checkout=branch,
                # Use a copy to prevent changes propagating among tests while test suite is running
                extra_context=extra_context.copy(),
                # Store replay data in the Briefcase template cache instead of ~/.cookiecutter_replay
                default_config={"replay_dir": str(self.template_cache_path(".replay"))},
            )
        except subprocess.CalledProcessError as e:
            # Computer is offline
            # status code == 128 - certificate validation error.
            raise NetworkFailure("clone template repository") from e
        except cookiecutter_exceptions.RepositoryNotFound as e:
            # Either the template path is invalid,
            # or it isn't a cookiecutter template (i.e., no cookiecutter.json)
            raise InvalidTemplateRepository(template) from e
        except cookiecutter_exceptions.RepositoryCloneFailed as e:
            # Branch does not exist.
            raise TemplateUnsupportedVersion(branch) from e

    def generate_template(
        self,
        template: str | None,
        branch: str | None,
        output_path: str | Path,
        extra_context: dict[str, str],
    ) -> None:
        # If a branch wasn't supplied through the --template-branch argument,
        # use the branch derived from the Briefcase version
        version = Version(briefcase.__version__)
        if branch is None:
            template_branch = f"v{version.base_version}"
        else:
            template_branch = branch

        extra_context = extra_context.copy()
        # Additional context that can be used for the Briefcase template pyproject.toml
        # header to include the version of Briefcase as well as the source of the template.
        extra_context.update(
            {
                "template_source": template,
                "template_branch": template_branch,
                "briefcase_version": str(version),
            }
        )

        try:
            self.logger.info(
                f"Using app template: {template}, branch {template_branch}"
            )
            # Unroll the new app template
            self._generate_template(
                template=template,
                branch=template_branch,
                output_path=output_path,
                extra_context=extra_context,
            )
        except TemplateUnsupportedVersion:
            # Only use the main template if we're on a development branch of briefcase
            # and the user didn't explicitly specify which branch to use.
            if version.dev is None or branch is not None:
                raise

            # Development branches can use the main template.
            self.logger.info(
                f"Template branch {template_branch} not found; falling back to development template"
            )

            extra_context["template_branch"] = "main"
            self._generate_template(
                template=template,
                branch="main",
                output_path=output_path,
                extra_context=extra_context,
            )



from __future__ import annotations

import concurrent.futures
import itertools
import os
import plistlib
import re
import subprocess
import time
from contextlib import suppress
from pathlib import Path
from signal import SIGTERM

from briefcase.config import AppConfig
from briefcase.console import select_option
from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.subprocess import get_process_id_by_command, is_process_dead
from briefcase.integrations.xcode import XcodeCliTools, get_identities
from briefcase.platforms.macOS.filters import macOS_log_clean_filter
from briefcase.platforms.macOS.utils import AppPackagesMergeMixin

try:
    import dmgbuild
except ImportError:  # pragma: no-cover-if-is-macos
    # On non-macOS platforms, dmgbuild won't be installed.
    # Allow the plugin to be loaded; raise an error when tools are verified.
    dmgbuild = None


DEFAULT_OUTPUT_FORMAT = "app"

ADHOC_IDENTITY_NAME = (
    "Ad-hoc identity. The resulting package will run but cannot be re-distributed."
)


class SigningIdentity:
    def __init__(self, id="-", name=None):
        """A wrapper around the various forms of an Apple signing identity."""
        self.id = id
        if self.id == "-":
            self.team_id = None
            self.name = ADHOC_IDENTITY_NAME
        else:
            self.name = name
            self.team_id = self.team_id_from_name(name)

    @classmethod
    def team_id_from_name(cls, name):
        try:
            return re.match(r".*\(([\dA-Z]*)\)", name)[1]
        except TypeError:
            raise BriefcaseCommandError(
                f"Couldn't extract Team ID from signing identity {name!r}"
            )

    @property
    def is_adhoc(self):
        """Is this the adhoc identity?"""
        return self.id == "-"

    def __repr__(self):
        if self.is_adhoc:
            return "<AdhocSigningIdentity>"
        else:
            return f"<SigningIdentity id={self.id}>"

    def __eq__(self, other):
        return isinstance(other, SigningIdentity) and self.id == other.id


class macOSMixin:
    platform = "macOS"
    supported_host_os = {"Darwin"}
    supported_host_os_reason = "macOS applications can only be built on macOS."


class macOSCreateMixin(AppPackagesMergeMixin):
    hidden_app_properties = {"permission", "entitlement"}

    def _install_app_requirements(
        self,
        app: AppConfig,
        requires: list[str],
        app_packages_path: Path,
    ):
        if getattr(app, "universal_build", True):
            # Perform the initial install targeting the current platform
            host_app_packages_path = (
                self.bundle_path(app) / f"app_packages.{self.tools.host_arch}"
            )
            super()._install_app_requirements(
                app,
                requires=requires,
                app_packages_path=host_app_packages_path,
            )

            # Find all the packages with binary components.
            # We can ignore any -universal2 packages; they're already fat.
            binary_packages = self.find_binary_packages(
                host_app_packages_path,
                universal_suffix="_universal2",
            )

            # Now install dependencies for the architecture that isn't the host architecture.
            other_arch = {
                "arm64": "x86_64",
                "x86_64": "arm64",
            }[self.tools.host_arch]

            # Create a temporary folder targeting the other platform
            other_app_packages_path = (
                self.bundle_path(app) / f"app_packages.{other_arch}"
            )
            if other_app_packages_path.is_dir():
                self.tools.shutil.rmtree(other_app_packages_path)
            self.tools.os.mkdir(other_app_packages_path)

            if binary_packages:
                with self.input.wait_bar(
                    f"Installing binary app requirements for {other_arch}..."
                ):
                    self._pip_install(
                        app,
                        app_packages_path=other_app_packages_path,
                        pip_args=[
                            "--no-deps",
                            "--only-binary",
                            ":all:",
                        ]
                        + [
                            f"{package}=={version}"
                            for package, version in binary_packages
                        ],
                        install_hint=f"""

If an {other_arch} wheel has not been published for one or more of your requirements,
you must compile those wheels yourself, or build a non-universal app by setting:

    universal_build = False

in the macOS configuration section of your pyproject.toml.
""",
                        env={
                            "PYTHONPATH": str(
                                self.support_path(app)
                                / "platform-site"
                                / f"macosx.{other_arch}"
                            )
                        },
                    )
            else:
                self.logger.info("All packages are pure Python, or universal.")

            # If given the option of a single architecture binary or a universal2 binary,
            # pip will install the single platform binary. However, a common situation on
            # macOS is for there to be an x86_64 binary and a universal2 binary. This means
            # you only get a universal2 binary in the "other" install pass. This then causes
            # problems with merging, because the "other" binary contains a copy of the
            # architecture that the "host" platform provides.
            #
            # To avoid this - ensure that the libraries in the app packages for the "other"
            # arch are all thin.
            #
            # This doesn't matter if it happens the other way around - if the "host" arch
            # installs a universal binary, then the "other" arch won't be asked to install
            # a binary at all.
            self.thin_app_packages(other_app_packages_path, arch=other_arch)

            # Merge the binaries
            self.merge_app_packages(
                target_app_packages=app_packages_path,
                sources=[host_app_packages_path, other_app_packages_path],
            )
        else:
            # If we're not building a universal binary, we can do a single install pass
            # directly into the app_packages folder.
            super()._install_app_requirements(
                app,
                requires=requires,
                app_packages_path=app_packages_path,
            )

            # Since we're only targeting 1 architecture, we can strip any universal
            # libraries down to just the host architecture.
            self.thin_app_packages(app_packages_path, arch=self.tools.host_arch)

    def permissions_context(self, app: AppConfig, cross_platform: dict[str, str]):
        """Additional template context for permissions.

        :param app: The config object for the app
        :param cross_platform: The dictionary of known cross-platform permission
            definitions.
        :returns: The template context describing permissions for the app.
        """
        # The info.plist entries for the app
        info = {}

        # Default entitlements for all macOS apps
        entitlements = {
            "com.apple.security.cs.allow-unsigned-executable-memory": True,
            "com.apple.security.cs.disable-library-validation": True,
        }

        if cross_platform["camera"]:
            entitlements["com.apple.security.device.camera"] = True
            info["NSCameraUsageDescription"] = cross_platform["camera"]
        if cross_platform["microphone"]:
            entitlements["com.apple.security.device.microphone"] = True
            info["NSMicrophoneUsageDescription"] = cross_platform["microphone"]

        if cross_platform["background_location"]:
            info["NSLocationUsageDescription"] = cross_platform["background_location"]
            entitlements["com.apple.security.personal-information.location"] = True
        elif cross_platform["fine_location"]:
            info["NSLocationUsageDescription"] = cross_platform["fine_location"]
            entitlements["com.apple.security.personal-information.location"] = True
        elif cross_platform["coarse_location"]:
            info["NSLocationUsageDescription"] = cross_platform["coarse_location"]
            entitlements["com.apple.security.personal-information.location"] = True

        if cross_platform["photo_library"]:
            info["NSPhotoLibraryUsageDescription"] = cross_platform["photo_library"]
            entitlements["com.apple.security.personal-information.photo_library"] = True

        # Override any info and entitlement definitions with the platform specific definitions
        info.update(getattr(app, "info", {}))
        entitlements.update(getattr(app, "entitlement", {}))

        return {
            "info": info,
            "entitlements": entitlements,
        }


class macOSRunMixin:
    def run_app(
        self,
        app: AppConfig,
        test_mode: bool,
        passthrough: list[str],
        **kwargs,
    ):
        """Start the application.

        :param app: The config object for the app
        :param test_mode: Boolean; Is the app running in test mode?
        :param passthrough: The list of arguments to pass to the app
        """
        # Console apps must operate in non-streaming mode so that console input can
        # be handled correctly. However, if we're in test mode, we *must* stream so
        # that we can see the test exit sentinel
        if app.console_app and not test_mode:
            self.run_console_app(
                app,
                passthrough=passthrough,
                **kwargs,
            )
        else:
            self.run_gui_app(
                app,
                test_mode=test_mode,
                passthrough=passthrough,
                **kwargs,
            )

    def run_console_app(
        self,
        app: AppConfig,
        passthrough: list[str],
        **kwargs,
    ):
        """Start the console application.

        :param app: The config object for the app
        :param passthrough: The list of arguments to pass to the app
        """
        try:
            kwargs = self._prepare_app_kwargs(app=app, test_mode=False)

            # Start the app directly
            self.logger.info("=" * 75)
            self.tools.subprocess.run(
                [self.binary_path(app) / "Contents" / "MacOS" / f"{app.formal_name}"]
                + (passthrough if passthrough else []),
                cwd=self.tools.home_path,
                check=True,
                stream_output=False,
                **kwargs,
            )

        except subprocess.CalledProcessError:
            # The command line app *could* returns an error code, which is entirely legal.
            # Ignore any subprocess error here.
            pass

    def run_gui_app(
        self,
        app: AppConfig,
        test_mode: bool,
        passthrough: list[str],
        **kwargs,
    ):
        """Start the GUI application.

        :param app: The config object for the app
        :param test_mode: Boolean; Is the app running in test mode?
        :param passthrough: The list of arguments to pass to the app
        """
        # Start log stream for the app.
        # Streaming the system log is... a mess. The system log contains a
        # *lot* of noise from other processes; even if you filter by
        # process, there's a lot of macOS-generated noise. It's very
        # difficult to extract just the "user generated" stdout/err log
        # messages.
        #
        # The following sets up a log stream filter that looks for:
        #  1. a log sender that matches that app binary; or,
        #  2. a log sender of libffi, and a process that matches the app binary.
        # Case (1) works for pre-Python 3.9 static linked binaries.
        # Case (2) works for Python 3.9+ dynamic linked binaries.
        # It's not enough to filter on *just* the processImagePath,
        # as the process will generate lots of system-level messages.
        # We can't filter on *just* the senderImagePath, because other
        # apps will generate log messages that would be caught by the filter.
        sender = os.fsdecode(self.binary_path(app) / "Contents/MacOS" / app.formal_name)
        log_popen = self.tools.subprocess.Popen(
            [
                "log",
                "stream",
                "--style",
                "compact",
                "--predicate",
                f'senderImagePath=="{sender}"'
                f' OR (processImagePath=="{sender}"'
                ' AND senderImagePath=="/usr/lib/libffi.dylib")',
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            bufsize=1,
        )

        # Wait for the log stream start up
        time.sleep(0.25)

        app_pid = None
        try:
            # Set up the log stream
            kwargs = self._prepare_app_kwargs(app=app, test_mode=test_mode)

            # Start the app in a way that lets us stream the logs
            self.tools.subprocess.run(
                # Force a new app to be launched
                ["open", "-n", self.binary_path(app)]
                + ((["--args"] + passthrough) if passthrough else []),
                cwd=self.tools.home_path,
                check=True,
                **kwargs,
            )

            # Find the App process ID so log streaming can exit when the app exits
            app_pid = get_process_id_by_command(
                command=str(self.binary_path(app)),
                logger=self.logger,
            )

            if app_pid is None:
                raise BriefcaseCommandError(
                    f"Unable to find process for app {app.app_name} to start log streaming."
                )

            # Stream the app logs.
            self._stream_app_logs(
                app,
                popen=log_popen,
                test_mode=test_mode,
                clean_filter=macOS_log_clean_filter,
                clean_output=True,
                stop_func=lambda: is_process_dead(app_pid),
                log_stream=True,
            )
        except subprocess.CalledProcessError:
            raise BriefcaseCommandError(f"Unable to start app {app.app_name}.")
        finally:
            # Ensure the App also terminates when exiting
            if app_pid:  # pragma: no-cover-if-is-py310
                with suppress(ProcessLookupError):
                    self.tools.os.kill(app_pid, SIGTERM)


def is_mach_o_binary(path):  # pragma: no-cover-if-is-windows
    """Determine if the file at the given path is a Mach-O binary.

    :param path: The path to check
    :returns: True if the file at the given location is a Mach-O binary.
    """
    # A binary is any file that is executable, or has a suffix from a known list
    if os.access(path, os.X_OK) or path.suffix.lower() in {".dylib", ".o", ".so", ""}:
        # File is a binary; read the file magic to determine if it's Mach-O.
        with path.open("rb") as f:
            magic = f.read(4)
            return magic in (
                b"\xCA\xFE\xBA\xBE",
                b"\xCF\xFA\xED\xFE",
                b"\xCE\xFA\xED\xFE",
                b"\xBE\xBA\xFE\xCA",
                b"\xFE\xED\xFA\xCF",
                b"\xFE\xED\xFA\xCE",
            )
    else:
        # Not a binary
        return False


class macOSSigningMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # External service APIs.
        # These are abstracted to enable testing without patching.
        self.get_identities = get_identities

    def entitlements_path(self, app: AppConfig):
        return self.bundle_path(app) / self.path_index(app, "entitlements_path")

    def select_identity(
        self,
        identity: str | None = None,
        app_identity: SigningIdentity | None = None,
    ) -> SigningIdentity:
        """Get the codesigning identity to use.

        This can be either an application codesigning identity, or an installer
        identity. An installer identity must be from the same Team ID as the application
        identity.

        :param identity: A pre-specified identity (either the 40-digit hex checksum, or
            the string name of the identity). If provided, it will be validated against
            the list of available identities to confirm that it is a valid codesigning
            identity.
        :param app_identity: The application signing identity to match when producing an
            installer identity. Only non-app signing identities that match the team ID
            for the ``app_identity`` will be presented as options. Omit this value to
            select an app signing identity.
        :returns: The final identity to use
        """
        # Obtain the valid codesigning identities. These are the identities that could
        # be used for app signing.
        identities = self.get_identities(self.tools, "codesigning")

        if app_identity:
            ident_type = "installer"
            ident_option = "--installer-identity"
            # There's no way to explicitly request a list of installer identities. As a
            # workaround, get the full, unfiltered list of all signing identities; then
            # remove any identity that:
            # 1. From a different team to the provided app identity; or
            # 2. Appears on the list of app signing identities.
            app_identities = identities
            identities = {
                key: name
                for key, name in self.get_identities(self.tools).items()
                if SigningIdentity.team_id_from_name(name) == app_identity.team_id
                and key not in app_identities
            }

            if not identities:
                raise BriefcaseCommandError(
                    f"No installer signing identities for team {app_identity.team_id} could be found."
                )
        else:
            ident_type = "application"
            ident_option = "--identity"

            # App signing also allows for the adhoc identity
            identities["-"] = ADHOC_IDENTITY_NAME

        if identity:
            try:
                # Try to look up the identity as a hex checksum
                identity_name = identities[identity]
                return SigningIdentity(id=identity, name=identity_name)
            except KeyError as e:
                # Try to look up the identity as readable name
                try:
                    reverse_lookup = {name: ident for ident, name in identities.items()}
                    identity_id = reverse_lookup[identity]
                    return SigningIdentity(id=identity_id, name=identity)
                except KeyError:
                    # Not found as an ID or name
                    raise BriefcaseCommandError(
                        f"Invalid {ident_type} signing identity {identity}"
                    ) from e

        self.input.prompt()
        self.input.prompt(f"Select {ident_type} signing identity to use:")
        self.input.prompt()
        identity = select_option(identities, input=self.input)
        identity_name = identities[identity]
        if identity == "-":
            self.logger.info(
                f"""
In the future, you could specify this signing identity by running:

    $ briefcase {self.command} macOS --adhoc-sign

"""
            )
        else:
            self.logger.info(
                f"""
In the future, you could specify this signing identity by running:

    $ briefcase {self.command} macOS {ident_option} {identity}

or

    $ briefcase {self.command} macOS {ident_option} "{identity_name}"

"""
            )

        return SigningIdentity(id=identity, name=identity_name)

    def sign_file(
        self,
        path: Path,
        identity: SigningIdentity,
        entitlements: Path | None = None,
    ):
        """Code sign a file.

        :param path: The path to the file to sign.
        :param identity: The code signing identity to use.
        :param entitlements: The path to the entitlements file to use.
        """
        options = "runtime" if not identity.is_adhoc else None
        process_command = ["codesign", path, "--sign", identity.id, "--force"]

        if entitlements:
            process_command.append("--entitlements")
            process_command.append(entitlements)
        if options:
            process_command.append("--options")
            process_command.append(options)

        self.logger.verbose(f"Signing {Path(path).relative_to(self.base_path)}")

        try:
            self.tools.subprocess.run(
                process_command,
                stderr=subprocess.PIPE,
                check=True,
            )
        except subprocess.CalledProcessError as e:
            errors = e.stderr
            if any(
                msg in errors
                for msg in [
                    # File has a signature matching the Mach-O magic,
                    # but isn't actually a Mach-O binary
                    "unsupported format for signature",
                    # A folder named ``.framework`, but not actually a macOS Framework`
                    "bundle format unrecognized, invalid, or unsuitable",
                ]
            ):
                # We should not be signing this in the first place
                self.logger.verbose(
                    f"... {Path(path).relative_to(self.base_path)} does not require a signature"
                )
                return
            else:
                raise BriefcaseCommandError(f"Unable to code sign {path}.")

    def sign_app(self, app: AppConfig, identity: SigningIdentity):
        """Sign an entire app with a specific identity.

        :param app: The app to sign
        :param identity: The signing identity to use
        """
        bundle_path = self.binary_path(app)
        resources_path = bundle_path / "Contents/Resources"
        frameworks_path = bundle_path / "Contents/Frameworks"

        sign_targets = []

        for folder in (resources_path, frameworks_path):
            # Sign all Mach-O executable objects
            sign_targets.extend(
                path
                for path in folder.rglob("*")
                if not path.is_dir() and is_mach_o_binary(path)
            )

            # Sign all embedded frameworks
            sign_targets.extend(folder.rglob("*.framework"))

            # Sign all embedded app objects
            sign_targets.extend(folder.rglob("*.app"))

        # Sign the bundle path itself
        sign_targets.append(bundle_path)

        # Run signing through a ThreadPoolExecutor so that they run in parallel.
        # However, we need to ensure that objects are signed from the inside out
        # (i.e., a folder must be signed *after* all it's contents has been
        # signed). To do this, we sort the list of signing targets in reverse
        # lexicographic order, and then group all the signing targets by parent.
        # This sorts all the signable files into folders; and sign all files in
        # a folder before sorting the next group. This ensures that longer paths
        # are signed first, and all files in a folder are signed before the
        # folder is signed.
        #
        # NOTE: We are relying on the fact that the final iteration order
        # produced by groupby() reflects the order in which groups are found in
        # the input data. The documentation for groupby() says that a new break
        # is created every time a new group is found in the input data; sorting
        # the input in reverse order ensures that only one group is found per folder,
        # and that the deepest folder is found first.
        progress_bar = self.input.progress_bar()
        task_id = progress_bar.add_task("Signing App", total=len(sign_targets))
        with progress_bar:
            for _, names in itertools.groupby(
                sorted(sign_targets, reverse=True),
                lambda name: name.parent,
            ):
                with concurrent.futures.ThreadPoolExecutor() as executor:
                    futures = []
                    for path in names:
                        future = executor.submit(
                            self.sign_file,
                            path,
                            entitlements=self.entitlements_path(app),
                            identity=identity,
                        )
                        futures.append(future)
                    for future in concurrent.futures.as_completed(futures):
                        progress_bar.update(task_id, advance=1)
                        if future.exception():
                            raise future.exception()


class macOSPackageMixin(macOSSigningMixin):
    ADHOC_SIGN_HELP = (
        "Perform ad-hoc signing on the app. "
        "The app will only run on this machine; it cannot be redistributed to others"
    )
    IDENTITY_HELP = (
        "The code signing identity to use; either the 40-digit hex "
        "checksum, or the full name of the identity"
    )

    @property
    def packaging_formats(self):
        return ["zip", "dmg", "pkg"]

    @property
    def default_packaging_format(self):
        # The default changes depending on whether the app is a console app or a GUI app
        return None

    def distribution_path(self, app):
        if app.packaging_format == "zip":
            return self.dist_path / f"{app.formal_name}-{app.version}.app.zip"
        elif app.packaging_format == "pkg":
            return self.dist_path / f"{app.formal_name}-{app.version}.pkg"
        else:
            return self.dist_path / f"{app.formal_name}-{app.version}.dmg"

    def add_options(self, parser):
        super().add_options(parser)

        # --no-sign-installer and --installer-identity are mutually exclusive
        installer_signing_group = parser.add_mutually_exclusive_group()
        installer_signing_group.add_argument(
            "--no-sign-installer",
            dest="sign_installer",
            help="Do not sign the installer. Ignored unless using PKG format",
            action="store_false",
        )
        installer_signing_group.add_argument(
            "--installer-identity",
            dest="installer_identity",
            help=(
                "The code signing identity to use for signing the installer. "
                "Ignored unless using PKG format"
            ),
            required=False,
        )

        # We use store_const:False rather than store_false so that the
        # "unspecified" value is None, rather than True, allowing for
        # a "default behavior" interpretation when `--adhoc-sign` is specified
        parser.add_argument(
            "--no-notarize",
            dest="notarize_app",
            action="store_const",
            const=False,
            help="Disable notarization for the app",
        )

    def verify_tools(self):
        # Require the Xcode command line tools.
        XcodeCliTools.verify(tools=self.tools)

        # Verify superclass tools *after* xcode. This ensures we get the
        # git check *after* the xcode check.
        super().verify_tools()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # External service APIs.
        # These are abstracted to enable testing without patching.
        self.dmgbuild = dmgbuild

    def verify_app(self, app):
        super().verify_app(app)

        if app.console_app:
            if app.packaging_format is None:
                app.packaging_format = "pkg"
            elif app.packaging_format != "pkg":
                raise BriefcaseCommandError(
                    "macOS console apps must be distributed in PKG format."
                )
        elif app.packaging_format is None:
            app.packaging_format = "dmg"

    def notarize(self, filename, identity: SigningIdentity):
        """Notarize a file.

        Submits the file to Apple for notarization; if successful, staples the
        notarization result onto the file.

        If the file is a .app, it will be archived as a .zip for submission purposes.

        :param filename: The file to notarize.
        :param identity: The code signing identity to use
        """
        try:
            if filename.suffix == ".app":
                # Archive the app into a zip.
                with self.input.wait_bar(f"Archiving {filename.name}..."):
                    archive_filename = filename.parent / "archive.zip"
                    self.tools.shutil.make_archive(
                        archive_filename.with_suffix(""),
                        format="zip",
                        root_dir=filename.parent,
                        base_dir=filename.name,
                    )
            elif filename.suffix in {".dmg", ".pkg"}:
                archive_filename = filename
            else:
                archive_filename = filename
                raise RuntimeError(
                    f"Don't know how to notarize a file of type {filename.suffix}"
                )

            profile = f"briefcase-macOS-{identity.team_id}"
            submitted = False
            store_credentials = False
            while not submitted:
                if store_credentials:
                    if not self.input.enabled:
                        raise BriefcaseCommandError(
                            f"""
The keychain does not contain credentials for the profile {profile}.
You can store these credentials by invoking:

    $ xcrun notarytool store-credentials --team-id {identity.team_id} profile

"""
                        )

                    self.logger.warning(
                        """
The notarization process uses credentials stored on your system Keychain.
You need to do this once for each signing certificate you use.

The credentials are authenticated and stored using your Apple ID, using
an app-specific Apple ID password. To generate an app-specific Apple ID
password:

  1. Sign into https://appleid.apple.com;
  2. In the 'Sign-in and Security' section, click 'App-Specific Passwords';
  3. Click on the '+' icon. You will need to provide an identifying name
     for the password. You can pick any name that makes sense to you - the
     name is only there so you can identify passwords. 'Briefcase' would be
     one possible name.
  4. Record the password somewhere safe.
"""
                    )
                    try:
                        self.tools.subprocess.run(
                            [
                                "xcrun",
                                "notarytool",
                                "store-credentials",
                                "--team-id",
                                identity.team_id,
                                profile,
                            ],
                            check=True,
                            stream_output=False,  # Command reads from stdin.
                        )
                    except subprocess.CalledProcessError as e:
                        raise BriefcaseCommandError(
                            f"Unable to store credentials for team ID {identity.team_id}."
                        ) from e

                # Attempt the notarization
                try:
                    self.logger.info()
                    self.tools.subprocess.run(
                        [
                            "xcrun",
                            "notarytool",
                            "submit",
                            archive_filename,
                            "--keychain-profile",
                            profile,
                            "--wait",
                        ],
                        check=True,
                    )
                    submitted = True
                except subprocess.CalledProcessError as e:
                    # Error when submitting for notarization.
                    # A return code of 69 (nice) indicates an issue with the
                    # keychain profile. If store_credentials is already True,
                    # then we've already tried to store them, so call the attempt
                    # a fail
                    if e.returncode == 69 and not store_credentials:
                        store_credentials = True
                    else:
                        raise BriefcaseCommandError(
                            f"""\
Unable to submit {filename.relative_to(self.base_path)} for notarization.
To find the cause of this failure, get the submission ID by running:

    xcrun notarytool history

Then run:

    xcrun notarytool log <submission-id>

to generate a full log of the error.
"""
                        ) from e
        finally:
            # Clean up house; we don't need the archive anymore.
            if archive_filename != filename:
                self.tools.os.unlink(archive_filename)

        try:
            self.logger.info()
            self.logger.info(
                f"Stapling notarization onto {filename.relative_to(self.base_path)}..."
            )
            self.tools.subprocess.run(
                ["xcrun", "stapler", "staple", filename],
                check=True,
            )
        except subprocess.CalledProcessError:
            raise BriefcaseCommandError(
                f"Unable to staple notarization onto {filename.relative_to(self.base_path)}."
            )

    def package_app(
        self,
        app: AppConfig,
        notarize_app=None,
        identity=None,
        adhoc_sign=False,
        sign_installer=True,
        installer_identity=None,
        **kwargs,
    ):
        """Package an app bundle.

        :param app: The application to package
        :param notarize_app: Should the app be notarized? Default: ``True`` if the app
            has been signed with a real identity; ``False`` if the app is unsigned, or
            an ad-hoc signing identity has been used.
        :param identity: The signing identity to use to sign the app. This can be either
            the 40-digit hex checksum, or the string name of the identity. If
            unspecified, the user will be prompted for an app signing identity. Ignored
            if ``adhoc_sign`` is ``True``.
        :param adhoc_sign: If ``True``, code will be signed with ad-hoc identity of "-",
            and the resulting app will not be re-distributable.
        :param sign_installer: Should the installer be signed? Ignored unless the
            packaging format is ``pkg``.
        :param installer_identity: The signing identity to use when signing the
            installer. Ignored unless the packaging format is ``pkg``.
        """
        self.logger.info("Signing app...", prefix=app.app_name)
        if adhoc_sign:
            identity = SigningIdentity()
        else:
            identity = self.select_identity(identity=identity)

        if identity.is_adhoc:
            if notarize_app:
                raise BriefcaseCommandError(
                    "Can't notarize an app with an ad-hoc signing identity"
                )
            self.logger.warning(
                """
*************************************************************************
** WARNING: Signing with an ad-hoc identity                            **
*************************************************************************

    This app is being signed with an ad-hoc identity. The resulting
    app will run on this computer, but will not run on anyone else's
    computer.

    To generate an app that can be distributed to others, you must
    obtain an application distribution certificate from Apple, and
    select the developer identity associated with that certificate
    when running 'briefcase package'.

*************************************************************************

"""
            )
            self.logger.info("Signing app with ad-hoc identity...")
        else:
            # If we're signing, and notarization isn't explicitly disabled,
            # notarize by default.
            if notarize_app is None:
                notarize_app = True

            self.logger.info(f"Signing app with identity {identity.name}...")

        self.sign_app(app=app, identity=identity)

        if app.packaging_format == "zip":
            self.package_zip(
                app,
                notarize_app=notarize_app,
                identity=identity,
            )

        elif app.packaging_format == "pkg":
            # If the user has indicated they want to sign the installer (the default),
            # and the signing identity for the app *isn't* the adhoc identity, select an
            # identity for signing the installer.
            if sign_installer and not identity.is_adhoc:
                installer_identity = self.select_identity(
                    identity=installer_identity,
                    app_identity=identity,
                )
            else:
                installer_identity = None

            self.package_pkg(
                app,
                notarize_app=notarize_app,
                identity=identity,
                installer_identity=installer_identity,
            )

        else:  # Default packaging format is DMG
            self.package_dmg(
                app,
                notarize_app=notarize_app,
                identity=identity,
            )

    def package_zip(
        self,
        app: AppConfig,
        notarize_app: bool,
        identity: SigningIdentity,
    ):
        """Package an .app bundle in a zip file."""
        dist_path: Path = self.distribution_path(app)

        if notarize_app:
            self.logger.info(
                f"Notarizing app using team ID {identity.team_id}...",
                prefix=app.app_name,
            )
            self.notarize(self.binary_path(app), identity=identity)

        with self.input.wait_bar(f"Archiving {dist_path.name}..."):
            self.tools.shutil.make_archive(
                dist_path.with_suffix(""),
                format="zip",
                root_dir=self.binary_path(app).parent,
                base_dir=self.binary_path(app).name,
            )

    def package_pkg(
        self,
        app: AppConfig,
        notarize_app: bool,
        identity: SigningIdentity,
        installer_identity: SigningIdentity | None,
    ):
        """Package the app as an installer."""
        dist_path: Path = self.distribution_path(app)

        self.logger.info("Building PKG...", prefix=app.app_name)

        installer_path = self.bundle_path(app) / "installer"

        with self.input.wait_bar("Installing license..."):
            license_file = self.base_path / "LICENSE"
            if license_file.is_file():
                (installer_path / "resources").mkdir(exist_ok=True)
                self.tools.shutil.copy(
                    license_file,
                    installer_path / "resources/LICENSE",
                )
            else:
                raise BriefcaseCommandError(
                    """\
Your project does not contain a LICENSE file.

Create a file named `LICENSE` in the same directory as your `pyproject.toml`
with your app's licensing terms.
"""
                )

        # pkgbuild's default behavior is to make "relocatable" installs, which means
        # that if you've ever run the app, the installer will default to updating *that*
        # version, rather than putting it in the location that the installer specifies.
        # This means if you've ever used `briefcase run`, that will be the install
        # location of the "installed" app. To work around this, you have to provide a
        # plist file - but that requires providing a "root" folder that *only* contains
        # the products you want to install. So - we need to copy the built app to a
        # "clean" packaging location.
        with self.input.wait_bar("Copying app into products folder..."):
            installed_app_path = installer_path / "root" / self.binary_path(app).name
            if installed_app_path.exists():
                self.tools.shutil.rmtree(installed_app_path)
            self.tools.shutil.copytree(self.binary_path(app), installed_app_path)

        components_plist_path = self.bundle_path(app) / "installer/components.plist"

        with self.input.wait_bar("Writing component manifest..."):
            with components_plist_path.open("wb") as components_plist:
                plistlib.dump(
                    [
                        {
                            "BundleHasStrictIdentifier": True,
                            "BundleIsRelocatable": False,
                            "BundleIsVersionChecked": True,
                            "BundleOverwriteAction": "upgrade",
                            "RootRelativeBundlePath": self.binary_path(app).name,
                        }
                    ],
                    components_plist,
                )

        # Console apps are installed in /Library/Formal Name, and include the
        # post-install scripts. Normal apps are installed in /Applications, and don't
        # include the scripts.
        if app.console_app:
            install_args = [
                "--install-location",
                f"/Library/{app.formal_name}",
                "--scripts",
                installer_path / "scripts",
            ]
        else:
            install_args = ["--install-location", "/Applications"]

        with self.input.wait_bar("Building app package..."):
            installer_packages_path = installer_path / "packages"
            if installer_packages_path.exists():
                self.tools.shutil.rmtree(installer_packages_path)
            installer_packages_path.mkdir()

            self.tools.subprocess.run(
                [
                    "pkgbuild",
                    "--root",
                    installer_path / "root",
                    "--component-plist",
                    components_plist_path,
                ]
                + install_args
                + [
                    installer_packages_path / f"{app.app_name}.pkg",
                ],
                check=True,
            )

        # Build package
        with self.input.wait_bar(f"Building {dist_path.name}..."):
            if installer_identity:
                signing_options = ["--sign", installer_identity.id]
            else:
                signing_options = []

            self.tools.subprocess.run(
                [
                    "productbuild",
                    "--distribution",
                    installer_path / "Distribution.xml",
                    "--package-path",
                    installer_path / "packages",
                    "--resources",
                    installer_path / "resources",
                ]
                + signing_options
                + [
                    dist_path,
                ],
                check=True,
            )

        if notarize_app:
            self.logger.info(
                f"Notarizing PKG with team ID {installer_identity.team_id}...",
                prefix=app.app_name,
            )
            self.notarize(dist_path, identity=installer_identity)

    def package_dmg(
        self,
        app: AppConfig,
        notarize_app: bool,
        identity: SigningIdentity,
    ):
        """Package an app as a DMG installer."""
        dist_path: Path = self.distribution_path(app)
        self.logger.info("Building DMG...", prefix=app.app_name)

        with self.input.wait_bar(f"Building {dist_path.name}..."):
            dmg_settings = {
                "files": [os.fsdecode(self.binary_path(app))],
                "symlinks": {"Applications": "/Applications"},
                "icon_locations": {
                    f"{app.formal_name}.app": (75, 75),
                    "Applications": (225, 75),
                },
                "window_rect": ((600, 600), (350, 150)),
                "icon_size": 64,
                "text_size": 12,
            }

            try:
                icon_filename = self.base_path / f"{app.installer_icon}.icns"
                if not icon_filename.exists():
                    self.logger.warning(
                        f"Can't find {app.installer_icon}.icns to use as DMG installer icon"
                    )
                    raise AttributeError()
            except AttributeError:
                # No installer icon specified. Fall back to the app icon
                if app.icon:
                    icon_filename = self.base_path / f"{app.icon}.icns"
                    if not icon_filename.exists():
                        self.logger.warning(
                            f"Can't find {app.icon}.icns to use as fallback DMG installer icon"
                        )
                        icon_filename = None
                else:
                    # No app icon specified either
                    icon_filename = None

            if icon_filename:
                dmg_settings["icon"] = os.fsdecode(icon_filename)

            try:
                image_filename = self.base_path / f"{app.installer_background}.png"
                if image_filename.exists():
                    dmg_settings["background"] = os.fsdecode(image_filename)
                else:
                    self.logger.warning(
                        f"Can't find {app.installer_background}.png to use as DMG background"
                    )
            except AttributeError:
                # No installer background image provided
                pass

            self.dmgbuild.build_dmg(
                filename=os.fsdecode(dist_path),
                volume_name=f"{app.formal_name} {app.version}",
                settings=dmg_settings,
            )

        self.sign_file(dist_path, identity=identity)

        if notarize_app:
            self.logger.info(
                f"Notarizing DMG with team ID {identity.team_id}...",
                prefix=app.app_name,
            )
            self.notarize(dist_path, identity=identity)



from unittest.mock import MagicMock

from ...utils import PartialMatchString


def test_default_without_http_protocol(convert_command, monkeypatch):
    mock_input_text = MagicMock()
    monkeypatch.setattr(convert_command, "input_text", mock_input_text)

    convert_command.input_bundle("http://www.some_url.no", "test-app-name", None)
    mock_input_text.assert_called_once_with(
        intro=PartialMatchString("no.some_url.www.test-app-name"),
        variable="bundle identifier",
        default="no.some_url.www",
        validator=convert_command.validate_bundle,
        override_value=None,
    )


def test_default_without_https_protocol(convert_command, monkeypatch):
    mock_input_text = MagicMock()
    monkeypatch.setattr(convert_command, "input_text", mock_input_text)

    convert_command.input_bundle("https://www.some_url.no", "test-app-name", None)
    mock_input_text.assert_called_once_with(
        intro=PartialMatchString("no.some_url.www.test-app-name"),
        variable="bundle identifier",
        default="no.some_url.www",
        validator=convert_command.validate_bundle,
        override_value=None,
    )


def test_override_is_used(convert_command):
    assert (
        convert_command.input_bundle("", "test-app-name", "com.overridden.project")
        == "com.overridden.project"
    )


def test_prompted_bundle(convert_command):
    """You can type in the bundle name."""
    convert_command.input.values = ["com.some.project"]
    assert convert_command.input_bundle("", "test-app-name", None) == "com.some.project"



import subprocess
from pathlib import Path

import pytest

from briefcase.console import LogLevel
from briefcase.exceptions import BriefcaseCommandError

from ...utils import create_file


@pytest.mark.parametrize("verbose", [True, False])
def test_lipo_dylib(dummy_command, verbose, tmp_path, capsys):
    """A binary library can be merged with lipo."""
    if verbose:
        dummy_command.logger.verbosity = LogLevel.VERBOSE

    # Create 3 source binaries.
    create_file(tmp_path / "source-1/path/to/file.dylib", "dylib-1")
    create_file(tmp_path / "source-2/path/to/file.dylib", "dylib-2")
    create_file(tmp_path / "source-3/path/to/file.dylib", "dylib-3")

    # Merge the libraries
    dummy_command.lipo_dylib(
        Path("path/to/file.dylib"),
        tmp_path / "target",
        [
            tmp_path / "source-1",
            tmp_path / "source-2",
            tmp_path / "source-3",
        ],
    )

    # Lipo was invoked with three sources
    dummy_command.tools.subprocess.run.assert_called_once_with(
        [
            "lipo",
            "-create",
            "-output",
            tmp_path / "target/path/to/file.dylib",
            tmp_path / "source-1/path/to/file.dylib",
            tmp_path / "source-2/path/to/file.dylib",
            tmp_path / "source-3/path/to/file.dylib",
        ],
        check=True,
    )

    # The target directory exists.
    assert (tmp_path / "target/path/to").is_dir()

    # Output only happens if in debug mode
    output = capsys.readouterr().out.split("\n")
    assert len(output) == (2 if verbose else 1)


@pytest.mark.parametrize("verbose", [True, False])
def test_lipo_dylib_partial(dummy_command, verbose, tmp_path, capsys):
    """If a source doesn't have the library, it isn't included in the merge."""
    if verbose:
        dummy_command.logger.verbosity = LogLevel.VERBOSE

    # Create 2 source binaries. Source-2 doesn't have the binary.
    create_file(tmp_path / "source-1/path/to/file.dylib", "dylib-1")
    create_file(tmp_path / "source-3/path/to/file.dylib", "dylib-3")

    # Merge the libraries
    dummy_command.lipo_dylib(
        Path("path/to/file.dylib"),
        tmp_path / "target",
        [
            tmp_path / "source-1",
            tmp_path / "source-3",
        ],
    )

    # Lipo was invoked with three sources
    dummy_command.tools.subprocess.run.assert_called_once_with(
        [
            "lipo",
            "-create",
            "-output",
            tmp_path / "target/path/to/file.dylib",
            tmp_path / "source-1/path/to/file.dylib",
            tmp_path / "source-3/path/to/file.dylib",
        ],
        check=True,
    )

    # The target directory exists.
    assert (tmp_path / "target/path/to").is_dir()

    # Output only happens if in debug mode
    output = capsys.readouterr().out.split("\n")
    assert len(output) == (2 if verbose else 1)


@pytest.mark.parametrize("verbose", [True, False])
def test_lipo_dylib_merge_error(dummy_command, verbose, tmp_path, capsys):
    """If the merge process fails, an exception is raised."""
    if verbose:
        dummy_command.logger.verbosity = LogLevel.VERBOSE

    # Create 3 source binaries.
    create_file(tmp_path / "source-1/path/to/file.dylib", "dylib-1")
    create_file(tmp_path / "source-2/path/to/file.dylib", "dylib-2")
    create_file(tmp_path / "source-3/path/to/file.dylib", "dylib-3")

    # lipo raises an exception.
    dummy_command.tools.subprocess.run.side_effect = subprocess.CalledProcessError(
        returncode=-1, cmd="lipo..."
    )

    # Merge the libraries. This raises an exception:
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to create fat library for path[/\\]to[/\\]file\.dylib",
    ):
        dummy_command.lipo_dylib(
            Path("path/to/file.dylib"),
            tmp_path / "target",
            [
                tmp_path / "source-1",
                tmp_path / "source-2",
                tmp_path / "source-3",
            ],
        )

    # Lipo was invoked with three sources
    dummy_command.tools.subprocess.run.assert_called_once_with(
        [
            "lipo",
            "-create",
            "-output",
            tmp_path / "target/path/to/file.dylib",
            tmp_path / "source-1/path/to/file.dylib",
            tmp_path / "source-2/path/to/file.dylib",
            tmp_path / "source-3/path/to/file.dylib",
        ],
        check=True,
    )

    # The target directory exists.
    assert (tmp_path / "target/path/to").is_dir()

    # Output only happens if in debug mode
    output = capsys.readouterr().out.split("\n")
    assert len(output) == (2 if verbose else 1)



import pytest

from briefcase.exceptions import UnsupportedHostError


@pytest.mark.parametrize("host_os", ["Darwin", "Linux", "Windows"])
def test_default_os_support(base_command, host_os):
    base_command.tools.host_os = host_os
    base_command.verify_host()


def test_unsupported_os(base_command):
    base_command.tools.host_os = "WeirdOS"
    with pytest.raises(UnsupportedHostError, match="This command is not supported on"):
        base_command.verify_host()



import pytest

from briefcase.exceptions import UpgradeToolError

from .conftest import (
    DummyManagedTool1,
    DummyManagedTool2,
    DummyManagedTool3,
    DummyNotInstalledManagedTool,
    DummyTool,
    DummyUnManagedManagedTool,
)


def test_list_tools(upgrade_command, mock_tool_registry, capsys):
    """The tools for upgrade can be listed."""
    upgrade_command(tool_list=[], list_tools=True)

    # Tools that are *not* relevant to this upgrade call are not verified
    DummyTool.verify.assert_not_called()

    # Tools that *are* relevant to this upgrade call are verified
    DummyUnManagedManagedTool.verify.assert_called_once_with(
        tools=upgrade_command.tools, install=False
    )
    DummyNotInstalledManagedTool.verify.assert_called_once_with(
        tools=upgrade_command.tools, install=False
    )
    DummyManagedTool1.verify.assert_called_once_with(
        tools=upgrade_command.tools, install=False
    )
    DummyManagedTool2.verify.assert_called_once_with(
        tools=upgrade_command.tools, install=False
    )
    DummyManagedTool3.verify.assert_called_once_with(
        tools=upgrade_command.tools, install=False
    )

    assert capsys.readouterr().out == (
        "\n"
        "[upgrade] Briefcase is managing the following tools:\n"
        " - Managed Dummy Tool 1 (managed_1)\n"
        " - Managed Dummy Tool 2 (managed_2)\n"
        " - Managed Dummy Tool 3 (managed_3)\n"
    )


def test_list_specific_tools(upgrade_command, mock_tool_registry, capsys):
    """If a list of tools is provided, only those are listed."""
    upgrade_command(tool_list=["managed_1", "managed_2"], list_tools=True)

    # Tools that are *not* relevant to this upgrade call are not verified
    DummyTool.verify.assert_not_called()
    DummyUnManagedManagedTool.verify.assert_not_called()
    DummyNotInstalledManagedTool.verify.assert_not_called()
    DummyManagedTool3.verify.assert_not_called()

    # Tools that *are* relevant to this upgrade call are verified
    DummyManagedTool1.verify.assert_called_once_with(
        tools=upgrade_command.tools, install=False
    )
    DummyManagedTool2.verify.assert_called_once_with(
        tools=upgrade_command.tools, install=False
    )

    assert capsys.readouterr().out == (
        "\n"
        "[upgrade] Briefcase is managing the following tools:\n"
        " - Managed Dummy Tool 1 (managed_1)\n"
        " - Managed Dummy Tool 2 (managed_2)\n"
    )


def test_upgrade_tools(upgrade_command, mock_tool_registry, capsys):
    """All managed tools can be upgraded."""
    upgrade_command(tool_list=[])

    # Tools that are *not* relevant to this upgrade call are not verified
    DummyTool.verify.assert_not_called()

    # Tools that *are* relevant to this upgrade call are verified
    DummyUnManagedManagedTool.verify.assert_called_once_with(
        tools=upgrade_command.tools, install=False
    )
    DummyNotInstalledManagedTool.verify.assert_called_once_with(
        tools=upgrade_command.tools, install=False
    )
    DummyManagedTool1.verify.assert_called_once_with(
        tools=upgrade_command.tools, install=False
    )
    DummyManagedTool2.verify.assert_called_once_with(
        tools=upgrade_command.tools, install=False
    )
    DummyManagedTool3.verify.assert_called_once_with(
        tools=upgrade_command.tools, install=False
    )

    # Tools that are *not* relevant to this upgrade call are not upgraded
    assert upgrade_command.tools.unmanaged_managed.actions == []

    # Tools that *are* relevant to this upgrade call are upgraded
    assert upgrade_command.tools.managed_1.actions == [
        "exists",
        "uninstall",
        "install",
    ]
    assert upgrade_command.tools.managed_2.actions == [
        "exists",
        "uninstall",
        "install",
    ]
    assert upgrade_command.tools.managed_3.actions == [
        "exists",
        "uninstall",
        "install",
    ]

    assert capsys.readouterr().out == (
        "\n"
        "[upgrade] Briefcase will upgrade the following tools:\n"
        " - Managed Dummy Tool 1 (managed_1)\n"
        " - Managed Dummy Tool 2 (managed_2)\n"
        " - Managed Dummy Tool 3 (managed_3)\n"
        "\n"
        "[managed_1] Upgrading Managed Dummy Tool 1...\n"
        "\n"
        "[managed_2] Upgrading Managed Dummy Tool 2...\n"
        "\n"
        "[managed_3] Upgrading Managed Dummy Tool 3...\n"
    )


def test_upgrade_specific_tools(upgrade_command, mock_tool_registry, capsys):
    """If a list of tools is provided, only those are listed."""
    upgrade_command(tool_list=["managed_1", "managed_2"])

    # Tools that are *not* relevant to this upgrade call are not verified
    DummyTool.verify.assert_not_called()
    DummyUnManagedManagedTool.verify.assert_not_called()
    DummyNotInstalledManagedTool.verify.assert_not_called()
    DummyManagedTool3.verify.assert_not_called()

    # Tools that *are* relevant to this upgrade call are verified
    DummyManagedTool1.verify.assert_called_once_with(
        tools=upgrade_command.tools, install=False
    )
    DummyManagedTool2.verify.assert_called_once_with(
        tools=upgrade_command.tools, install=False
    )

    # Tools that *are* relevant to this upgrade call are upgraded
    assert upgrade_command.tools.managed_1.actions == [
        "exists",
        "uninstall",
        "install",
    ]
    assert upgrade_command.tools.managed_2.actions == [
        "exists",
        "uninstall",
        "install",
    ]

    assert capsys.readouterr().out == (
        "\n"
        "[upgrade] Briefcase will upgrade the following tools:\n"
        " - Managed Dummy Tool 1 (managed_1)\n"
        " - Managed Dummy Tool 2 (managed_2)\n"
        "\n"
        "[managed_1] Upgrading Managed Dummy Tool 1...\n"
        "\n"
        "[managed_2] Upgrading Managed Dummy Tool 2...\n"
    )


def test_upgrade_no_tools(upgrade_command, mock_no_managed_tool_registry, capsys):
    """If no tools are being managed, a message is returned."""
    upgrade_command(tool_list=[])

    # Tools that are *not* relevant to this upgrade call are not verified
    DummyTool.verify.assert_not_called()

    # Tools that *are* relevant to this upgrade call are verified
    DummyUnManagedManagedTool.verify.assert_called_once_with(
        tools=upgrade_command.tools, install=False
    )
    DummyNotInstalledManagedTool.verify.assert_called_once_with(
        tools=upgrade_command.tools, install=False
    )

    assert capsys.readouterr().out == "Briefcase is not managing any tools.\n"


def test_upgrade_unmanaged_tools(upgrade_command, mock_tool_registry, capsys):
    """If only unmanaged tools are requested to upgrade, error is raised."""
    with pytest.raises(
        UpgradeToolError,
        match="Briefcase is not managing not_installed, unmanaged, unmanaged_managed.",
    ):
        upgrade_command(tool_list=["unmanaged", "unmanaged_managed", "not_installed"])

    # Tools that are *not* relevant to this upgrade call are not verified
    DummyTool.verify.assert_not_called()
    DummyManagedTool1.verify.assert_not_called()
    DummyManagedTool2.verify.assert_not_called()
    DummyManagedTool3.verify.assert_not_called()

    # Tools that *are* relevant to this upgrade call are verified
    DummyUnManagedManagedTool.verify.assert_called_once_with(
        tools=upgrade_command.tools, install=False
    )
    DummyNotInstalledManagedTool.verify.assert_called_once_with(
        tools=upgrade_command.tools, install=False
    )

    assert capsys.readouterr().out == ""


def test_upgrade_mixed_tools(upgrade_command, mock_tool_registry, capsys):
    """If managed and unmanaged tools are requested to upgrade, a warning is shown and
    the upgrade continues."""
    upgrade_command(
        tool_list=[
            "managed_1",
            "managed_2",
            "unmanaged",
            "unmanaged_managed",
            "not_installed",
        ]
    )

    # Tools that are *not* relevant to this upgrade call are not verified
    DummyTool.verify.assert_not_called()
    DummyManagedTool3.verify.assert_not_called()

    # Tools that *are* relevant to this upgrade call are verified
    DummyUnManagedManagedTool.verify.assert_called_once_with(
        tools=upgrade_command.tools, install=False
    )
    DummyNotInstalledManagedTool.verify.assert_called_once_with(
        tools=upgrade_command.tools, install=False
    )

    # Tools that are *not* relevant to this upgrade call are not upgraded
    assert upgrade_command.tools.unmanaged_managed.actions == []

    # Tools that *are* relevant to this upgrade call are upgraded
    assert upgrade_command.tools.managed_1.actions == [
        "exists",
        "uninstall",
        "install",
    ]
    assert upgrade_command.tools.managed_2.actions == [
        "exists",
        "uninstall",
        "install",
    ]

    assert capsys.readouterr().out == (
        "Briefcase is not managing not_installed, unmanaged, unmanaged_managed.\n"
        "\n"
        "[upgrade] Briefcase will upgrade the following tools:\n"
        " - Managed Dummy Tool 1 (managed_1)\n"
        " - Managed Dummy Tool 2 (managed_2)\n"
        "\n"
        "[managed_1] Upgrading Managed Dummy Tool 1...\n"
        "\n"
        "[managed_2] Upgrading Managed Dummy Tool 2...\n"
    )


def test_unknown_tool(upgrade_command, mock_tool_registry, capsys):
    """An upgrade attempt for an unknown tool raises an error."""

    with pytest.raises(
        UpgradeToolError,
        match="Briefcase does not know how to manage unknown_tool_1, unknown_tool_2.",
    ):
        upgrade_command(tool_list=["managed_1", "unknown_tool_1", "unknown_tool_2"])

    # Tools that are *not* relevant to this upgrade call are not verified
    DummyTool.verify.assert_not_called()
    DummyManagedTool1.verify.assert_not_called()
    DummyManagedTool2.verify.assert_not_called()
    DummyManagedTool3.verify.assert_not_called()
    DummyUnManagedManagedTool.verify.assert_not_called()
    DummyNotInstalledManagedTool.verify.assert_not_called()

    assert capsys.readouterr().out == ""



import shutil
import sys
from unittest.mock import MagicMock

import pytest


@pytest.fixture(scope="module")
def custom_packing_format():
    shutil.register_unpack_format(
        "custom_packing",
        [".archive", ".archive.ext"],
        lambda x, y: True,
    )
    yield
    shutil.unregister_unpack_format("custom_packing")


@pytest.mark.parametrize(
    "filename, outcome",
    [
        ("filename.tar", True),
        ("filename.zip", True),
        ("filename.archive", True),
        ("filename.part.archive", True),
        ("filename.archive.ext", True),
        ("filename.part.archive.ext", True),
        ("filename.part.archive.ext", True),
        ("filename", False),
        ("filename.doc", False),
        ("filename.archive.doc", False),
        ("filename.archive.ext.doc", False),
    ],
)
@pytest.mark.usefixtures("custom_packing_format")
def test_is_archive(mock_tools, filename, outcome, monkeypatch):
    """Archive filenames are properly detected."""
    assert mock_tools.file.is_archive(filename) is outcome


def test_unpack_archive(mock_tools):
    """Archive unpacking is deferred correctly for an arbitrary archive."""
    mock_tools.shutil = MagicMock(spec=shutil)

    mock_tools.file.unpack_archive(
        "test_archive",
        extract_dir="destination",
    )

    mock_tools.shutil.unpack_archive.assert_called_once_with(
        filename="test_archive",
        extract_dir="destination",
        **({"filter": "data"} if sys.version_info >= (3, 12) else {}),
    )


def test_unpack_archive_kwargs(mock_tools):
    """Archive unpacking is deferred correctly with kwargs."""
    mock_tools.shutil = MagicMock(spec=shutil)

    mock_tools.file.unpack_archive(
        "test_archive",
        extract_dir="destination",
        extra_arg="arg this",
    )

    mock_tools.shutil.unpack_archive.assert_called_once_with(
        filename="test_archive",
        extract_dir="destination",
        **(
            {"filter": "data", "extra_arg": "arg this"}
            if sys.version_info >= (3, 12)
            else {"extra_arg": "arg this"}
        ),
    )


def test_unpack_archive_override_filter(mock_tools):
    """Archive unpacking is deferred correctly while overriding `filter`."""
    mock_tools.shutil = MagicMock(spec=shutil)

    mock_tools.file.unpack_archive(
        "test_archive",
        extract_dir="destination",
        filter="onlycatpics",
        extra_arg="arg this",
    )

    mock_tools.shutil.unpack_archive.assert_called_once_with(
        filename="test_archive",
        extract_dir="destination",
        filter="onlycatpics",
        extra_arg="arg this",
    )


def test_unpack_zip_archive(mock_tools):
    """Archive unpacking is deferred correctly for ZIP archives."""
    mock_tools.shutil = MagicMock(spec=shutil)

    mock_tools.file.unpack_archive(
        "test_archive.zip",
        extract_dir="destination",
    )

    mock_tools.shutil.unpack_archive.assert_called_once_with(
        filename="test_archive.zip",
        extract_dir="destination",
    )


def test_unpack_zip_archive_kwargs(mock_tools):
    """Archive unpacking is deferred correctly for ZIP archives with kwargs."""
    mock_tools.shutil = MagicMock(spec=shutil)

    mock_tools.file.unpack_archive(
        "test_archive.zip",
        extract_dir="destination",
        extra_arg="arg this",
    )

    mock_tools.shutil.unpack_archive.assert_called_once_with(
        filename="test_archive.zip",
        extract_dir="destination",
        extra_arg="arg this",
    )



from unittest.mock import MagicMock

import pytest

from briefcase.exceptions import BriefcaseCommandError, InvalidDeviceError
from briefcase.integrations.android_sdk import (
    ADB,
    AndroidDeviceNotAuthorized,
    AndroidSDK,
)


@pytest.fixture
def android_sdk(android_sdk, tmp_path) -> AndroidSDK:
    android_sdk.devices = MagicMock(
        return_value={
            "041234567892009a": {
                "name": "Unknown device (not authorized for development)",
                "authorized": False,
            },
            "KABCDABCDA1513": {
                "name": "Kogan Agora 9",
                "authorized": True,
            },
            "emulator-5554": {
                "name": "Android SDK built for x86",
                "authorized": True,
            },
        }
    )

    android_sdk.emulators = MagicMock(
        return_value=[
            "runningEmulator",
            "idleEmulator",
        ]
    )

    # Set up an ADB for each device.
    def mock_adb(device_id):
        adb = MagicMock(spec_set=ADB)
        if device_id == "emulator-5554":
            adb.avd_name.return_value = "runningEmulator"
        else:
            adb.avd_name.return_value = None
        return adb

    android_sdk.adb = mock_adb

    return android_sdk


def test_explicit_device(mock_tools, android_sdk):
    """If the user explicitly names a physical device, it is returned."""

    # Select device with an explicit device ID
    device, name, avd = android_sdk.select_target_device("KABCDABCDA1513")

    # Physical running device, so no AVD
    assert device == "KABCDABCDA1513"
    assert name == "Kogan Agora 9 (KABCDABCDA1513)"
    assert avd is None

    # No input was requested
    assert mock_tools.input.prompts == []


def test_explicit_unauthorized_device(mock_tools, android_sdk):
    """If the user explicitly names an unauthorized physical device, an error is
    raised."""

    # Select unauthorized device with an explicit device ID
    with pytest.raises(AndroidDeviceNotAuthorized):
        android_sdk.select_target_device("041234567892009a")

    # No input was requested
    assert mock_tools.input.prompts == []


def test_explicit_running_emulator_by_id(mock_tools, android_sdk):
    """If the user explicitly names a running emulator by device ID, it is selected."""

    # Select emulator with an explicit device ID
    device, name, avd = android_sdk.select_target_device("emulator-5554")

    # Emulator is running, so there is a device ID
    assert device == "emulator-5554"
    assert name == "@runningEmulator (running emulator)"
    assert avd == "runningEmulator"

    # No input was requested
    assert mock_tools.input.prompts == []


def test_explicit_running_emulator_by_avd(mock_tools, android_sdk):
    """If the user explicitly names a running emulator by AVD, it is selected."""

    # Select emulator with an explicit device ID
    device, name, avd = android_sdk.select_target_device("@runningEmulator")

    # Emulator is running, so there is a device ID
    assert device == "emulator-5554"
    assert name == "@runningEmulator (running emulator)"
    assert avd == "runningEmulator"

    # No input was requested
    assert mock_tools.input.prompts == []


def test_explicit_idle_emulator(mock_tools, android_sdk):
    """If the user explicitly names an idle emulator by AVD, it is selected."""

    # Select emulator with an explicit device ID
    device, name, avd = android_sdk.select_target_device("@idleEmulator")

    # Emulator is not running, so no device ID
    assert device is None
    assert name == "@idleEmulator (emulator)"
    assert avd == "idleEmulator"

    # No input was requested
    assert mock_tools.input.prompts == []


def test_explicit_invalid_device(mock_tools, android_sdk):
    """If the user explicitly names a non-existet device, an error is raised."""

    # Select emulator with an invalid device ID
    with pytest.raises(InvalidDeviceError):
        device, name, avd = android_sdk.select_target_device("deadbeefcafe")

    # No input was requested
    assert mock_tools.input.prompts == []


def test_explicit_invalid_avd(mock_tools, android_sdk):
    """If the user explicitly names a non-existent device, an error is raised."""

    # Select emulator with an invalid AVD
    with pytest.raises(InvalidDeviceError):
        device, name, avd = android_sdk.select_target_device("@invalidEmulator")

    # No input was requested
    assert mock_tools.input.prompts == []


def test_select_device(mock_tools, android_sdk, capsys):
    """If the user manually selects a physical device, details are returned."""
    # Mock the user input
    mock_tools.input.values = ["2"]

    # Run the selection with no pre-existing choice
    device, name, avd = android_sdk.select_target_device(None)

    # Physical running device, so no AVD
    assert device == "KABCDABCDA1513"
    assert name == "Kogan Agora 9 (KABCDABCDA1513)"
    assert avd is None

    # The user was asked to select a device
    assert len(mock_tools.input.prompts) == 1

    # A re-run prompt has been provided
    out = capsys.readouterr().out
    assert "briefcase run android -d KABCDABCDA1513" in out


def test_select_unauthorized_device(mock_tools, android_sdk):
    """If the user manually selects an unauthorized running device, an error is
    raised."""
    # Mock the user input
    mock_tools.input.values = ["3"]

    # Run the selection with no pre-existing choice
    with pytest.raises(AndroidDeviceNotAuthorized):
        android_sdk.select_target_device(None)


def test_select_running_emulator(mock_tools, android_sdk, capsys):
    """If the user manually selects a running emulator, details are returned."""
    # Mock the user input
    mock_tools.input.values = ["1"]

    # Run the selection with no pre-existing choice
    device, name, avd = android_sdk.select_target_device(None)

    # Emulator is running, so there is a device ID
    assert device == "emulator-5554"
    assert name == "@runningEmulator (running emulator)"
    assert avd == "runningEmulator"

    # A re-run prompt has been provided
    out = capsys.readouterr().out
    assert 'briefcase run android -d "@runningEmulator"' in out


def test_select_idle_emulator(mock_tools, android_sdk, capsys):
    """If the user manually selects a running device, details are returned."""
    # Mock the user input
    mock_tools.input.values = ["4"]

    # Run the selection with no pre-existing choice
    device, name, avd = android_sdk.select_target_device(None)

    # Emulator is not running, so no device ID
    assert device is None
    assert name == "@idleEmulator (emulator)"
    assert avd == "idleEmulator"

    # A re-run prompt has been provided
    out = capsys.readouterr().out
    assert 'briefcase run android -d "@idleEmulator"' in out


def test_select_create_emulator(mock_tools, android_sdk, capsys):
    """If the user manually selects a running device, details are returned."""
    # Mock the user input
    mock_tools.input.values = ["5"]

    # Run the selection with no pre-existing choice
    device, name, avd = android_sdk.select_target_device(None)

    # No device details
    assert device is None
    assert name is None
    assert avd is None

    # A re-run prompt has *not* been provided
    out = capsys.readouterr().out
    assert "In future, you could specify this device" not in out


def test_input_disabled(mock_tools, android_sdk):
    """If input has been disabled, and there are multiple simulators, an error is
    raised."""
    mock_tools.input.enabled = False

    # Run the selection with no pre-existing choice
    with pytest.raises(BriefcaseCommandError):
        android_sdk.select_target_device(None)

    # No input was requested
    assert mock_tools.input.prompts == []


def test_input_disabled_no_simulators(mock_tools, android_sdk):
    """If input has been disabled, and there are no simulators, 'create' is selected."""
    mock_tools.input.enabled = False

    # Remove all the devices and emulators
    android_sdk.devices = MagicMock(return_value={})
    android_sdk.emulators = MagicMock(return_value=[])

    # Run the selection with no pre-existing choice
    device, name, avd = android_sdk.select_target_device(None)

    # No device details
    assert device is None
    assert name is None
    assert avd is None

    # No input was requested
    assert mock_tools.input.prompts == []


def test_input_disabled_one_device(mock_tools, android_sdk):
    """If input has been disabled, and there is a single device, it is selected."""
    mock_tools.input.enabled = False

    # Set up a single device.
    android_sdk.devices = MagicMock(
        return_value={
            "KABCDABCDA1513": {
                "name": "Kogan Agora 9",
                "authorized": True,
            },
        }
    )
    android_sdk.emulators = MagicMock(return_value=[])

    # Run the selection with no pre-existing choice
    device, name, avd = android_sdk.select_target_device(None)

    # The only device is returned
    assert device == "KABCDABCDA1513"
    assert name == "Kogan Agora 9 (KABCDABCDA1513)"
    assert avd is None

    # No input was requested
    assert mock_tools.input.prompts == []


def test_explicit_new_device(android_sdk):
    """If the user provides a device configuration, that configuration is created."""
    android_sdk._create_emulator = MagicMock()

    # Select a target device by providing a device config
    device, name, avd = android_sdk.select_target_device('{"avd":"myDevice"}')

    # A request was made to create the emulator using the internal method.
    android_sdk._create_emulator.assert_called_once_with(avd="myDevice")

    assert device is None
    assert name == "@myDevice (emulator)"
    assert avd == "myDevice"


def test_explicit_new_device_existing(android_sdk):
    """If the user provides a device configuration, that configuration is created."""
    android_sdk._create_emulator = MagicMock()

    # Select a target device by providing a device config
    device, name, avd = android_sdk.select_target_device('{"avd":"idleEmulator"}')

    # The emulator already exists, so it won't be created.
    android_sdk._create_emulator.assert_not_called()

    assert device is None
    assert name == "@idleEmulator (emulator)"
    assert avd == "idleEmulator"


def test_explicit_new_device_full_spec(android_sdk):
    """If the user provides a device configuration, that configuration is created."""
    android_sdk._create_emulator = MagicMock()

    # Select a target device by providing a device config
    device, name, avd = android_sdk.select_target_device(
        '{"avd":"myDevice","device_type":"pixel","skin":"pixel_6",'
        '"system_image":"system-images;android-21;default;arm64-v8a"}'
    )

    # A request was made to create the emulator using the internal method.
    android_sdk._create_emulator.assert_called_once_with(
        avd="myDevice",
        device_type="pixel",
        skin="pixel_6",
        system_image="system-images;android-21;default;arm64-v8a",
    )

    assert device is None
    assert name == "@myDevice (emulator)"
    assert avd == "myDevice"


def test_explicit_new_device_bad_spec(android_sdk):
    """If the user provides a bad device configuration, an error is raised."""
    android_sdk._create_emulator = MagicMock()

    # Select a target device by providing a device config
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to create emulator with definition '{NOT A JSON SNIPPET}'",
    ):
        android_sdk.select_target_device("{NOT A JSON SNIPPET}")


def test_explicit_new_device_incomplete_spec(android_sdk):
    """If the user provides a device configuration but doesn't provide an AVD, an error
    is raised."""
    android_sdk._create_emulator = MagicMock()

    # Select a target device by providing a device config
    with pytest.raises(
        BriefcaseCommandError,
        match=r"No AVD provided for new device",
    ):
        android_sdk.select_target_device('{"device_type":"pixel","skin":"pixel_3a"}')


def test_explicit_new_device_unknown_spec(android_sdk):
    """If the user provides a device configuration with an unknown argument, an error is
    raised."""

    def mock_create_emulator(avd, device_type=None, skin=None, system_image=None):
        pass

    android_sdk._create_emulator = MagicMock(side_effect=mock_create_emulator)

    # Select a target device by providing a device config
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unknown device property 'color'",
    ):
        android_sdk.select_target_device(
            '{"avd":"myPhone","device_type":"pixel","color":"blue"}'
        )



import pytest

from briefcase.integrations.windows_sdk import WindowsSDK

from .test_WindowsSDK__verify import setup_winsdk_install


@pytest.mark.parametrize(
    "versions, expected",
    [
        ([], []),
        (["85.0.1"], ["85.0.1.0"]),
        # ensure the newest version is first
        (["85.0.1", "85.0.2", "85.0.3"], ["85.0.3.0", "85.0.2.0", "85.0.1.0"]),
        # ensure invalid versions are ignored
        (["85.0.1", "85.0.2", "86.0.3"], ["85.0.2.0", "85.0.1.0"]),
        # ensure no versions if all versions are invalid
        (["86.0.1", "86.0.2", "86.0.3"], []),
    ],
)
def test_sdk_versions_from_bin(tmp_path, versions, expected, monkeypatch):
    """Versions from SDK bin directory are properly found and sorted."""
    # Patch target SDK version
    monkeypatch.setattr(WindowsSDK, "SDK_VERSION", "85.0")
    monkeypatch.setattr(WindowsSDK, "SDK_MIN_VERSION", 0)

    sdk_path = tmp_path / "win_sdk"
    for version in versions:
        setup_winsdk_install(sdk_path.parent, version)

    assert WindowsSDK._sdk_versions_from_bin(sdk_path) == expected
